import sys

from pandajedi.jedicore.MsgWrapper import MsgWrapper
from .TaskSetupperBase import TaskSetupperBase

from pandaserver.dataservice import DataServiceUtils
from pandaserver.taskbuffer import EventServiceUtils, JobUtils

from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# task setup for ATLAS
class AtlasTaskSetupper (TaskSetupperBase):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        TaskSetupperBase.__init__(self,taskBufferIF,ddmIF)


    # main to setup task
    def doSetup(self,taskSpec,datasetToRegister,pandaJobs):
        # make logger
        tmpLog = MsgWrapper(logger,"< jediTaskID={0} >".format(taskSpec.jediTaskID))
        tmpLog.info('start label={0} taskType={1}'.format(taskSpec.prodSourceLabel,taskSpec.taskType))
        # returns
        retFatal    = self.SC_FATAL
        retTmpError = self.SC_FAILED
        retOK       = self.SC_SUCCEEDED
        try:
            # get DDM I/F
            ddmIF = self.ddmIF.getInterface(taskSpec.vo)
            # register datasets
            if datasetToRegister != [] or taskSpec.prodSourceLabel in ['user']:
                # prod vs anal
                userSetup = False
                if taskSpec.prodSourceLabel in ['user']:
                    userSetup = True
                    # collect datasetID to register datasets/containers just in case
                    for tmpPandaJob in pandaJobs:
                        if not tmpPandaJob.produceUnMerge():
                            for tmpFileSpec in tmpPandaJob.Files:
                                if tmpFileSpec.type in ['output','log']:
                                    if tmpFileSpec.datasetID not in datasetToRegister:
                                        datasetToRegister.append(tmpFileSpec.datasetID)
                tmpLog.info('datasetToRegister={0}'.format(str(datasetToRegister)))
                # get site mapper
                siteMapper = self.taskBufferIF.getSiteMapper()

                # loop over all datasets
                avDatasetList = []
                cnDatasetMap  = {}
                for datasetID in datasetToRegister:
                    # get output and log datasets
                    tmpLog.info('getting datasetSpec with datasetID={0}'.format(datasetID))
                    tmpStat,datasetSpec = self.taskBufferIF.getDatasetWithID_JEDI(taskSpec.jediTaskID,
                                                                                  datasetID)
                    if not tmpStat:
                        tmpLog.error('failed to get output and log datasets')
                        return retFatal
                    if datasetSpec.isPseudo():
                        tmpLog.info('skip pseudo dataset')
                        continue
                    # DDM backend
                    ddmBackEnd = taskSpec.getDdmBackEnd()
                    tmpLog.info('checking {0}'.format(datasetSpec.datasetName))
                    # check if dataset and container are available in DDM
                    for targetName in [datasetSpec.datasetName,datasetSpec.containerName]:
                        if targetName is None:
                            continue
                        if targetName not in avDatasetList:
                            # set lifetime
                            if targetName.startswith('panda'):
                                if datasetSpec.type == 'trn_log' and taskSpec.prodSourceLabel == 'managed':
                                    lifetime = 365
                                else:
                                    lifetime = 14
                            else:
                                lifetime = None
                            # check dataset/container in DDM
                            tmpList = ddmIF.listDatasets(targetName)
                            if tmpList == []:
                                # get location
                                location = None
                                locForRule = None
                                if targetName == datasetSpec.datasetName:
                                    # dataset
                                    if datasetSpec.site in ['',None]:
                                        if DataServiceUtils.getDistributedDestination(datasetSpec.storageToken) is not None:
                                            locForRule = datasetSpec.destination
                                        elif DataServiceUtils.getDestinationSE(datasetSpec.storageToken) is not None:
                                            location = DataServiceUtils.getDestinationSE(datasetSpec.storageToken)
                                        elif taskSpec.cloud is not None:
                                            # use T1 SE
                                            tmpT1Name = siteMapper.getCloud(taskSpec.cloud)['source']
                                            location = siteMapper.getDdmEndpoint(tmpT1Name, datasetSpec.storageToken,
                                                                                 taskSpec.prodSourceLabel,
                                                                                 JobUtils.translate_tasktype_to_jobtype(taskSpec.taskType))
                                    else:
                                        tmpLog.info('site={0} token={1}'.format(datasetSpec.site, datasetSpec.storageToken))
                                        location = siteMapper.getDdmEndpoint(datasetSpec.site,datasetSpec.storageToken,
                                                                             taskSpec.prodSourceLabel,
                                                                             JobUtils.translate_tasktype_to_jobtype(taskSpec.taskType))
                                if locForRule is None:
                                    locForRule = location
                                # set metadata
                                if taskSpec.prodSourceLabel in ['managed','test'] and targetName == datasetSpec.datasetName:
                                    metaData = {}
                                    metaData['task_id'] = taskSpec.jediTaskID
                                    if taskSpec.campaign not in [None,'']:
                                        metaData['campaign'] = taskSpec.campaign
                                    if datasetSpec.getTransient() is not None:
                                        metaData['transient'] = datasetSpec.getTransient()
                                else:
                                    metaData = None
                                # register dataset/container
                                tmpLog.info('registering {0} with location={1} backend={2} lifetime={3} meta={4}'.format(targetName,
                                                                                                                         location,
                                                                                                                         ddmBackEnd,
                                                                                                                         lifetime,
                                                                                                                         str(metaData)))
                                tmpStat = ddmIF.registerNewDataset(targetName,backEnd=ddmBackEnd,location=location,
                                                                   lifetime=lifetime,metaData=metaData)
                                if not tmpStat:
                                    tmpLog.error('failed to register {0}'.format(targetName))
                                    return retFatal
                                # procedures for user
                                if userSetup or DataServiceUtils.getDistributedDestination(datasetSpec.storageToken) is not None:
                                    # register location
                                    tmpToRegister = False
                                    if userSetup and targetName == datasetSpec.datasetName and datasetSpec.site not in ['',None]:
                                        if taskSpec.workingGroup:
                                            userName = taskSpec.workingGroup
                                        else:
                                            userName = taskSpec.userName
                                        grouping = None
                                        tmpToRegister = True
                                    elif DataServiceUtils.getDistributedDestination(datasetSpec.storageToken) is not None:
                                        userName = None
                                        grouping = 'NONE'
                                        tmpToRegister = True
                                    if tmpToRegister:
                                        activity = DataServiceUtils.getActivityForOut(taskSpec.prodSourceLabel)
                                        tmpLog.info('registering location={} lifetime={} days activity={} grouping={} '
                                                    'owner={}'.format(locForRule, lifetime, activity, grouping,
                                                                      userName))
                                        tmpStat = ddmIF.registerDatasetLocation(targetName,locForRule,owner=userName,
                                                                                lifetime=lifetime,backEnd=ddmBackEnd,
                                                                                activity=activity,grouping=grouping)
                                        if not tmpStat:
                                            tmpLog.error('failed to register location {0} for {1}'.format(locForRule,
                                                                                                          targetName))
                                            return retFatal
                                        # double copy
                                        if userSetup and datasetSpec.type == 'output':
                                            if datasetSpec.destination != datasetSpec.site:
                                                tmpLog.info('skip making double copy as destination={0} is not site={1}'.format(datasetSpec.destination,
                                                                                                                                datasetSpec.site))
                                            else:

                                                second_copy = True
                                                try:
                                                    if taskSpec.site:
                                                        panda_site = siteMapper.getSite(taskSpec.site)
                                                        if panda_site.catchall and 'skip_2nd_copy' in panda_site.catchall:
                                                            tmpLog.info('skip making double copy as specified in {0} catchall'.format(panda_site))
                                                            second_copy = False
                                                except Exception:
                                                    second_copy = True

                                                if second_copy:
                                                    locForDouble = '(type=SCRATCHDISK)\\notforextracopy=True'
                                                    tmpMsg  = 'registering double copy '
                                                    tmpMsg += 'location="{0}" lifetime={1}days activity={2} for dataset={3}'.format(locForDouble,lifetime,
                                                                                                                                    activity,targetName)
                                                    tmpLog.info(tmpMsg)
                                                    tmpStat = ddmIF.registerDatasetLocation(targetName,locForDouble,copies=2,owner=userName,
                                                                                            lifetime=lifetime,activity=activity,
                                                                                            grouping='NONE',weight='freespace',
                                                                                            ignore_availability=False)
                                                    if not tmpStat:
                                                        tmpLog.error('failed to register double copylocation {0} for {1}'.format(locForDouble,
                                                                                                                               targetName))
                                                        return retFatal
                                avDatasetList.append(targetName)
                            else:
                                tmpLog.info('{0} already registered'.format(targetName))
                    # check if dataset is in the container
                    if datasetSpec.containerName is not None and datasetSpec.containerName != datasetSpec.datasetName:
                        # get list of constituent datasets in the container
                        if datasetSpec.containerName not in cnDatasetMap:
                            cnDatasetMap[datasetSpec.containerName] = ddmIF.listDatasetsInContainer(datasetSpec.containerName)
                        # add dataset
                        if datasetSpec.datasetName not in cnDatasetMap[datasetSpec.containerName]:
                            tmpLog.info('adding {0} to {1}'.format(datasetSpec.datasetName,datasetSpec.containerName))
                            tmpStat = ddmIF.addDatasetsToContainer(datasetSpec.containerName,[datasetSpec.datasetName],
                                                                   backEnd=ddmBackEnd)
                            if not tmpStat:
                                tmpLog.error('failed to add {0} to {1}'.format(datasetSpec.datasetName,
                                                                               datasetSpec.containerName))
                                return retFatal
                            cnDatasetMap[datasetSpec.containerName].append(datasetSpec.datasetName)
                        else:
                            tmpLog.info('{0} already in {1}'.format(datasetSpec.datasetName,datasetSpec.containerName))
                    # update dataset
                    datasetSpec.status = 'registered'
                    self.taskBufferIF.updateDataset_JEDI(datasetSpec,{'jediTaskID':taskSpec.jediTaskID,
                                                                      'datasetID':datasetID})
            # register ES datasets
            if taskSpec.registerEsFiles():
                targetName = EventServiceUtils.getEsDatasetName(taskSpec.jediTaskID)
                location = None
                metaData = {}
                metaData['task_id'] = taskSpec.jediTaskID
                metaData['hidden']  = True
                tmpLog.info('registering ES dataset {0} with location={1} meta={2}'.format(targetName,
                                                                                           location,
                                                                                           str(metaData)))
                tmpStat = ddmIF.registerNewDataset(targetName,location=location,metaData=metaData,
                                                   resurrect=True)
                if not tmpStat:
                    tmpLog.error('failed to register ES dataset {0}'.format(targetName))
                    return retFatal
                # register rule
                location = 'type=DATADISK'
                activity = DataServiceUtils.getActivityForOut(taskSpec.prodSourceLabel)
                grouping = 'NONE'
                tmpLog.info('registering location={0} activity={1} grouping={2}'.format(location,
                                                                                        activity,
                                                                                        grouping))
                tmpStat = ddmIF.registerDatasetLocation(targetName,location,activity=activity,
                                                        grouping=grouping)
                if not tmpStat:
                    tmpLog.error('failed to register location {0} with {2} for {1}'.format(location,
                                                                                           targetName,
                                                                                           activity))
                    return retFatal
            # open datasets
            if taskSpec.prodSourceLabel in ['managed','test']:
                # get the list of output/log datasets
                outDatasetList = []
                for tmpPandaJob in pandaJobs:
                    for tmpFileSpec in tmpPandaJob.Files:
                        if tmpFileSpec.type in ['output','log']:
                            if tmpFileSpec.destinationDBlock not in outDatasetList:
                                outDatasetList.append(tmpFileSpec.destinationDBlock)
                # open datasets
                for outDataset in outDatasetList:
                    tmpLog.info('open {0}'.format(outDataset))
                    ddmIF.openDataset(outDataset)
                    # unset lifetime
                    ddmIF.setDatasetMetadata(outDataset,'lifetime',None)
            # return
            tmpLog.info('done')
            return retOK
        except Exception:
            errtype,errvalue = sys.exc_info()[:2]
            tmpLog.error('doSetup failed with {0}:{1}'.format(errtype.__name__,errvalue))
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retFatal



import sys
try:
    metaID = sys.argv[1]
except Exception:
    metaID = None
import json
import uuid
taskParamMap = {}
taskParamMap['taskName'] = str(uuid.uuid4())
taskParamMap['userName'] = 'pandasrv1'
taskParamMap['vo'] = 'atlas'
taskParamMap['taskPriority'] = 100
taskParamMap['architecture'] = 'i686-slc5-gcc43-opt'
taskParamMap['transUses'] = 'Atlas-17.2.7'
taskParamMap['transHome'] = 'AtlasProduction-17.2.8.10'
taskParamMap['transPath'] = 'Reco_trf.py'
taskParamMap['processingType'] = 'reco'
taskParamMap['prodSourceLabel'] = 'test'
taskParamMap['taskType'] = 'prod'
taskParamMap['workingGroup'] = 'AP_Higgs'
taskParamMap['coreCount'] = 1
taskParamMap['cloud'] = 'US'
logDatasetName = 'panda.jeditest.log.{0}'.format(uuid.uuid4())
taskParamMap['log'] = {'dataset': logDatasetName,
                       'type':'template',
                       'param_type':'log',
                       'token':'ATLASDATADISK',
                       'value':'{0}.${{SN}}.log.tgz'.format(logDatasetName)}
outDatasetName = 'panda.jeditest.NTUP_EMBLLDN.{0}'.format(uuid.uuid4())
taskParamMap['jobParameters'] = [
    {'type':'template',
     'param_type':'input',
     'value':'inputAODFile=${IN}',
     'dataset':'data12_8TeV.00214651.physics_Egamma.merge.AOD.f489_m1261',
     },
    {'type':'constant',
     'value': 'maxEvents=1000 RunNumber=213816 autoConfiguration=everything preExec="from BTagging.BTaggingFlags import BTaggingFlags;BTaggingFlags.CalibrationTag=\"BTagCalibALL-07-02\""'
     },
    {'type':'template',
     'param_type':'input',
     'value':'DBRelease=${DBR}',
     'dataset':'ddo.000001.Atlas.Ideal.DBRelease.v220701',
     'attributes':'repeat,nosplit',
     },
    {'type':'constant',
     'value':'AMITag=p1462'
     },
    {'type':'template',
     'param_type':'output',
     'token':'ATLASDATADISK',
     'value':'outputNTUP_EMBLLDNFile={0}.${{SN}}.pool.root'.format(outDatasetName),
     'dataset':outDatasetName,
     }
    ]

jonStr = json.dumps(taskParamMap)

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface  # noqa: E402

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()
tbIF.insertTaskParams_JEDI(taskParamMap['vo'],taskParamMap['prodSourceLabel'],taskParamMap['userName'],taskParamMap['taskName'],jonStr)



import json

from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin
from pandajedi.jedimsgprocessor.tape_carousel_msg_processor import TapeCarouselMsgProcPlugin
from pandajedi.jedimsgprocessor.hpo_msg_processor import HPOMsgProcPlugin
from pandajedi.jedimsgprocessor.processing_msg_processor import ProcessingMsgProcPlugin

from pandacommon.pandalogger import logger_utils


# logger
base_logger = logger_utils.setup_logger(__name__.split('.')[-1])


# Atlas iDDS message processing plugin, a bridge connect to other idds related message processing plugins
class AtlasIddsMsgProcPlugin(BaseMsgProcPlugin):

    def initialize(self):
        BaseMsgProcPlugin.initialize(self)
        self.plugin_TapeCarousel = TapeCarouselMsgProcPlugin()
        self.plugin_HPO = HPOMsgProcPlugin()
        self.plugin_Processing = ProcessingMsgProcPlugin()
        # for each plugin
        for _plugin in [self.plugin_TapeCarousel, self.plugin_HPO, self.plugin_Processing]:
            # initialize each
            _plugin.initialize()
            # use the same taskBuffer interface
            try:
                del _plugin.tbIF
            except Exception:
                pass
            _plugin.tbIF = self.tbIF

    def process(self, msg_obj):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, method_name='process')
        # start
        tmp_log.info('start')
        tmp_log.debug('sub_id={0} ; msg_id={1}'.format(msg_obj.sub_id, msg_obj.msg_id))
        # parse json
        try:
            msg_dict = json.loads(msg_obj.data)
        except Exception as e:
            err_str = 'failed to parse message json {2} , skipped. {0} : {1}'.format(e.__class__.__name__, e, msg_obj.data)
            tmp_log.error(err_str)
            raise
        # sanity check
        try:
            msg_type = msg_dict['msg_type']
        except Exception as e:
            err_str = 'failed to parse message object dict {2} , skipped. {0} : {1}'.format(e.__class__.__name__, e, msg_dict)
            tmp_log.error(err_str)
            raise
        # run different plugins according to message type
        try:
            if msg_type in ('file_stagein', 'collection_stagein', 'work_stagein'):
                self.plugin_TapeCarousel.process(msg_obj, decoded_data=msg_dict)
            elif msg_type in ('file_hyperparameteropt', 'collection_hyperparameteropt', 'work_hyperparameteropt'):
                self.plugin_HPO.process(msg_obj, decoded_data=msg_dict)
            elif msg_type in ('file_processing', 'collection_processing', 'work_processing'):
                self.plugin_Processing.process(msg_obj, decoded_data=msg_dict)
            else:
                # Asked by iDDS and message broker guys, JEDI needs to consume unknown types of messages and do nothing...
                warn_str = 'unknown msg_type : {0}'.format(msg_type)
                tmp_log.warning(warn_str)
        except Exception:
            raise
        # done
        tmp_log.info('done')



import re
import shlex
import random
import datetime
import traceback
from six import iteritems

from .TaskRefinerBase import TaskRefinerBase
from pandajedi.jedicore import JediException
from pandaserver.dataservice import DataServiceUtils


# brokerage for ATLAS production
class AtlasProdTaskRefiner (TaskRefinerBase):

    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TaskRefinerBase.__init__(self, taskBufferIF, ddmIF)



    # extract common parameters
    def extractCommon(self,jediTaskID,taskParamMap,workQueueMapper,splitRule):
        tmpLog = self.tmpLog
        # set ddmBackEnd
        if 'ddmBackEnd' not in taskParamMap:
            taskParamMap['ddmBackEnd'] = 'rucio'
        # get number of unprocessed events for event service
        autoEsConversion = False
        if 'esConvertible' in taskParamMap and taskParamMap['esConvertible'] is True:
            maxPrio = self.taskBufferIF.getConfigValue('taskrefiner', 'AES_MAXTASKPRIORITY', 'jedi', 'atlas')
            minPrio = self.taskBufferIF.getConfigValue('taskrefiner', 'AES_MINTASKPRIORITY', 'jedi', 'atlas')
            if maxPrio is not None and maxPrio < taskParamMap['taskPriority']:
                pass
            elif minPrio is not None and minPrio > taskParamMap['taskPriority']:
                pass
            else:
                # get threshold
                minNumEvents = self.taskBufferIF.getConfigValue('taskrefiner', 'AES_EVENTPOOLSIZE', 'jedi', 'atlas')
                maxPending = self.taskBufferIF.getConfigValue('taskrefiner', 'AES_MAXPENDING', 'jedi', 'atlas')
                nEvents, lastTaskTime, nPendingTasks = self.taskBufferIF.getNumUnprocessedEvents_JEDI(taskParamMap['vo'],
                                                                                                      taskParamMap['prodSourceLabel'],
                                                                                                      {'eventService': 1},
                                                                                                      {'gshare': 'Validation'})
                tmpStr = 'check for ES '
                tmpStr += 'tot_num_unprocessed_events_AES={0} target_num_events_AES={1} last_AES_task_time={2} '.format(nEvents,
                                                                                                                        minNumEvents,
                                                                                                                        lastTaskTime)
                tmpStr += 'num_pending_tasks_AES={0} max_pending_tasks_AES={1} '.format(nPendingTasks, maxPending)
                tmpLog.info(tmpStr)
                # not chane many tasks at once
                if lastTaskTime is None or (lastTaskTime < datetime.datetime.utcnow() - datetime.timedelta(minutes=5)):
                    if minNumEvents is not None and nEvents < minNumEvents and \
                            maxPending is not None and (maxPending is None or maxPending > nPendingTasks):
                        autoEsConversion = True
                        tmpLog.info('will be converted to AES unless it goes to pending')
        # add ES paramsters
        if ('esFraction' in taskParamMap and taskParamMap['esFraction'] > 0) or \
           ('esConvertible' in taskParamMap and taskParamMap['esConvertible'] is True):
            tmpStr  = '<PANDA_ES_ONLY>--eventService=True</PANDA_ES_ONLY>'
            taskParamMap['jobParameters'].append({'type':'constant',
                                                  'value':tmpStr})
        if ('esFraction' in taskParamMap and taskParamMap['esFraction'] > 0) or autoEsConversion:
            if 'nEventsPerWorker' not in taskParamMap and \
                    (('esFraction' in taskParamMap and taskParamMap['esFraction'] > random.random()) or autoEsConversion):
                taskParamMap['nEventsPerWorker'] = 1
                taskParamMap['registerEsFiles'] = True
                if 'nEsConsumers' not in taskParamMap:
                    tmpVal = self.taskBufferIF.getConfigValue('taskrefiner', 'AES_NESCONSUMERS', 'jedi', 'atlas')
                    if tmpVal is None:
                        tmpVal = 1
                    taskParamMap['nEsConsumers'] = tmpVal
                if 'nSitesPerJob' not in taskParamMap:
                    tmpVal = self.taskBufferIF.getConfigValue('taskrefiner', 'AES_NSITESPERJOB', 'jedi', 'atlas')
                    if tmpVal is not None:
                        taskParamMap['nSitesPerJob'] = tmpVal
                if 'mergeEsOnOS' not in taskParamMap:
                    taskParamMap['mergeEsOnOS'] = True
                if 'maxAttemptES' not in taskParamMap:
                    taskParamMap['maxAttemptES'] = 1
                if 'maxAttemptEsJob' not in taskParamMap:
                    taskParamMap['maxAttemptEsJob'] = 0
                if 'notDiscardEvents' not in taskParamMap:
                    taskParamMap['notDiscardEvents'] = True
                if 'decAttOnFailedES' not in taskParamMap:
                    taskParamMap['decAttOnFailedES'] = True
                taskParamMap['coreCount'] = 0
                taskParamMap['resurrectConsumers'] = True
        TaskRefinerBase.extractCommon(self,jediTaskID,taskParamMap,workQueueMapper,splitRule)



    # main
    def doRefine(self,jediTaskID,taskParamMap):
        # make logger
        tmpLog = self.tmpLog
        tmpLog.debug('start taskType={0}'.format(self.taskSpec.taskType))
        try:
            # basic refine
            self.doBasicRefine(taskParamMap)
            # set nosplit+repeat for DBR
            for datasetSpec in self.inSecDatasetSpecList:
                if DataServiceUtils.isDBR(datasetSpec.datasetName):
                    datasetSpec.attributes = 'repeat,nosplit'
            # enable consistency check
            if self.taskSpec.parent_tid not in [None,self.taskSpec.jediTaskID]:
                for datasetSpec in self.inMasterDatasetSpec:
                    if datasetSpec.isMaster() and datasetSpec.type == 'input':
                        datasetSpec.enableCheckConsistency()
            # append attempt number
            for tmpKey,tmpOutTemplateMapList in iteritems(self.outputTemplateMap):
                for tmpOutTemplateMap in tmpOutTemplateMapList:
                    outFileTemplate = tmpOutTemplateMap['filenameTemplate']
                    if re.search('\.\d+$',outFileTemplate) is None and not outFileTemplate.endswith('.panda.um'):
                        tmpOutTemplateMap['filenameTemplate'] = outFileTemplate + '.1'
            # extract input datatype
            datasetTypeListIn = []
            for datasetSpec in self.inMasterDatasetSpec+self.inSecDatasetSpecList:
                datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                if datasetType not in ['',None]:
                    datasetTypeListIn.append(datasetType)
            # extract datatype and set destination if nessesary
            datasetTypeList = []
            for datasetSpec in self.outDatasetSpecList:
                datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                if datasetType not in ['',None]:
                    datasetTypeList.append(datasetType)
            # set numThrottled to use the task throttling mechanism
            if 'noThrottle' not in taskParamMap:
                self.taskSpec.numThrottled = 0
            # set to register datasets
            self.taskSpec.setToRegisterDatasets()
            # set transient to parent datasets
            if self.taskSpec.processingType in ['merge'] and self.taskSpec.parent_tid not in [None,self.taskSpec.jediTaskID]:
                # get parent
                tmpStat,parentTaskSpec = self.taskBufferIF.getTaskDatasetsWithID_JEDI(self.taskSpec.parent_tid,None,False)
                if tmpStat and parentTaskSpec is not None:
                    # set transient to parent datasets
                    metaData = {'transient':True}
                    for datasetSpec in parentTaskSpec.datasetSpecList:
                        if datasetSpec.type in ['log','output']:
                            datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                            if datasetType not in ['',None] and datasetType in datasetTypeList and datasetType in datasetTypeListIn:
                                tmpLog.info('set metadata={0} to parent jediTaskID={1}:datasetID={2}:Name={3}'.format(str(metaData),
                                                                                                                      self.taskSpec.parent_tid,
                                                                                                                      datasetSpec.datasetID,
                                                                                                                      datasetSpec.datasetName))
                                for metadataName,metadaValue in iteritems(metaData):
                                    self.ddmIF.getInterface(self.taskSpec.vo).setDatasetMetadata(datasetSpec.datasetName,
                                                                                                 metadataName,metadaValue)
            # input prestaging
            if self.taskSpec.inputPreStaging():
                # set first contents feed flag
                self.taskSpec.set_first_contents_feed(True)
        except JediException.UnknownDatasetError as e:
            tmpLog.debug('in doRefine. {0}'.format(str(e)))
            raise e
        except Exception as e:
            tmpLog.error('doRefine failed with {0} {1}'.format(str(e), traceback.format_exc()))
            raise e
        tmpLog.debug('done')
        return self.SC_SUCCEEDED



    # insert string
    def insertString(self,paramName,tmpStr,origStr):
        items = shlex.split(origStr,posix=False)
        newStr = ''
        for item in items:
            if paramName not in item:
                newStr += item
            else:
                newStr += item[:-1]
                newStr += tmpStr
                newStr += item[-1]
            newStr += ' '
        return newStr



import re
import sys

from six import iteritems

from .PostProcessorBase import PostProcessorBase
from . import AtlasPostProcessorUtils

from pandaserver.dataservice import DataServiceUtils
from pandaserver.taskbuffer import EventServiceUtils


# post processor for ATLAS production
class AtlasProdPostProcessor (PostProcessorBase):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        PostProcessorBase.__init__(self,taskBufferIF,ddmIF)


    # main
    def doPostProcess(self,taskSpec,tmpLog):
        # pre-check
        try:
            tmpStat = self.doPreCheck(taskSpec,tmpLog)
            if tmpStat:
                return self.SC_SUCCEEDED
        except Exception:
            errtype,errvalue = sys.exc_info()[:2]
            tmpLog.error('doPreCheck failed with {0}:{1}'.format(errtype.__name__,errvalue))
            return self.SC_FATAL
        # get DDM I/F
        ddmIF = self.ddmIF.getInterface(taskSpec.vo)
        # loop over all datasets
        for datasetSpec in taskSpec.datasetSpecList:
            # skip pseudo output datasets
            if datasetSpec.type in ['output'] and datasetSpec.isPseudo():
                continue
            try:
                # remove wrong files
                if datasetSpec.type in ['output']:
                    # get successful files
                    okFiles = self.taskBufferIF.getSuccessfulFiles_JEDI(datasetSpec.jediTaskID,datasetSpec.datasetID)
                    if okFiles is None:
                        tmpLog.warning('failed to get successful files for {0}'.format(datasetSpec.datasetName))
                        return self.SC_FAILED
                    # get files in dataset
                    ddmFiles = ddmIF.getFilesInDataset(datasetSpec.datasetName,skipDuplicate=False,ignoreUnknown=True)
                    tmpLog.debug('datasetID={0}:Name={1} has {2} files in DB, {3} files in DDM'.format(datasetSpec.datasetID,
                                                                                                      datasetSpec.datasetName,
                                                                                                      len(okFiles),len(ddmFiles)))
                    # check all files
                    toDelete = []
                    for tmpGUID,attMap in iteritems(ddmFiles):
                        if attMap['lfn'] not in okFiles:
                            did = {'scope':attMap['scope'], 'name':attMap['lfn']}
                            toDelete.append(did)
                            tmpLog.debug('delete {0} from {1}'.format(attMap['lfn'],datasetSpec.datasetName))
                    # delete
                    if toDelete != []:
                        ddmIF.deleteFilesFromDataset(datasetSpec.datasetName,toDelete)
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                tmpLog.warning('failed to remove wrong files with {0}:{1}'.format(errtype.__name__,errvalue))
                return self.SC_FAILED
            try:
                # freeze output and log datasets
                if datasetSpec.type in ['output','log','trn_log']:
                    tmpLog.info('freezing datasetID={0}:Name={1}'.format(datasetSpec.datasetID,datasetSpec.datasetName))
                    ddmIF.freezeDataset(datasetSpec.datasetName,ignoreUnknown=True)
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                tmpLog.warning('failed to freeze datasets with {0}:{1}'.format(errtype.__name__,errvalue))
                return self.SC_FAILED
            try:
                # delete transient datasets
                if datasetSpec.type in ['trn_output']:
                    tmpLog.debug('deleting datasetID={0}:Name={1}'.format(datasetSpec.datasetID,datasetSpec.datasetName))
                    retStr = ddmIF.deleteDataset(datasetSpec.datasetName,False,ignoreUnknown=True)
                    tmpLog.info(retStr)
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                tmpLog.warning('failed to delete datasets with {0}:{1}'.format(errtype.__name__,errvalue))
        # check duplication
        if self.getFinalTaskStatus(taskSpec) in ['finished','done'] and taskSpec.gshare != 'Test':
            nDup = self.taskBufferIF.checkDuplication_JEDI(taskSpec.jediTaskID)
            tmpLog.debug('checked duplication with {0}'.format(nDup))
            if nDup > 0:
                errStr = 'paused since {0} duplication found'.format(nDup)
                taskSpec.oldStatus = self.getFinalTaskStatus(taskSpec)
                taskSpec.status = 'paused'
                taskSpec.setErrDiag(errStr)
                tmpLog.debug(errStr)
        # delete ES datasets
        if taskSpec.registerEsFiles():
            try:
                targetName = EventServiceUtils.getEsDatasetName(taskSpec.jediTaskID)
                tmpLog.debug('deleting ES dataset name={0}'.format(targetName))
                retStr = ddmIF.deleteDataset(targetName,False,ignoreUnknown=True)
                tmpLog.debug(retStr)
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                tmpLog.warning('failed to delete ES dataset with {0}:{1}'.format(errtype.__name__,errvalue))
        try:
            AtlasPostProcessorUtils.send_notification(self.taskBufferIF, ddmIF, taskSpec, tmpLog)
        except Exception as e:
            tmpLog.error('failed to talk to external system with {0}'.format(str(e)))
            return self.SC_FAILED
        try:
            self.doBasicPostProcess(taskSpec,tmpLog)
        except Exception:
            errtype,errvalue = sys.exc_info()[:2]
            tmpLog.error('doBasicPostProcess failed with {0}:{1}'.format(errtype.__name__,errvalue))
            return self.SC_FATAL
        return self.SC_SUCCEEDED



    # final procedure
    def doFinalProcedure(self,taskSpec,tmpLog):
        tmpLog.info('final procedure for status={0} processingType={1}'.format(taskSpec.status,
                                                                               taskSpec.processingType))
        if taskSpec.status in ['done','finished'] or \
                (taskSpec.status == 'paused' and taskSpec.oldStatus in ['done','finished']):
            trnLifeTime = 14*24*60*60
            trnLifeTimeMerge = 40*24*60*60
            ddmIF = self.ddmIF.getInterface(taskSpec.vo)
            # set lifetime to transient datasets
            metaData = {'lifetime':trnLifeTime}
            datasetTypeListI = set()
            datasetTypeListO = set()
            for datasetSpec in taskSpec.datasetSpecList:
                if datasetSpec.type in ['log','output']:
                    if datasetSpec.getTransient() is True:
                        tmpLog.debug('set metadata={0} to datasetID={1}:Name={2}'.format(str(metaData),
                                                                                        datasetSpec.datasetID,
                                                                                        datasetSpec.datasetName))
                        for metadataName,metadaValue in iteritems(metaData):
                            ddmIF.setDatasetMetadata(datasetSpec.datasetName,metadataName,metadaValue)
                # collect dataset types
                datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                if datasetType not in ['',None]:
                    if datasetSpec.type == 'input':
                        datasetTypeListI.add(datasetType)
                    elif datasetSpec.type == 'output':
                        datasetTypeListO.add(datasetType)
            # set lifetime to parent transient datasets
            if taskSpec.processingType in ['merge'] and \
                    (taskSpec.status == 'done' or \
                         (taskSpec.status == 'finished' and self.getFinalTaskStatus(taskSpec, checkParent=False) == 'done') or \
                         (taskSpec.status == 'paused' and \
                              (taskSpec.oldStatus == 'done' or \
                                   (taskSpec.oldStatus == 'finished' and self.getFinalTaskStatus(taskSpec, checkParent=False) == 'done')))):
                # get parent task
                if taskSpec.parent_tid not in [None,taskSpec.jediTaskID]:
                    # get parent
                    tmpStat,parentTaskSpec = self.taskBufferIF.getTaskDatasetsWithID_JEDI(taskSpec.parent_tid,None,False)
                    if tmpStat and parentTaskSpec is not None:
                        # set lifetime to parent datasets if they are transient
                        for datasetSpec in parentTaskSpec.datasetSpecList:
                            if datasetSpec.type in ['output']:
                                # check dataset type
                                datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                                if datasetType not in datasetTypeListI or datasetType not in datasetTypeListO:
                                    continue
                                metaData = {'lifetime': trnLifeTimeMerge}
                                tmpMetadata = ddmIF.getDatasetMetaData(datasetSpec.datasetName)
                                if tmpMetadata['transient'] is True:
                                    tmpLog.debug('set metadata={0} to parent jediTaskID={1}:datasetID={2}:Name={3}'.format(str(metaData),
                                                                                                                          taskSpec.parent_tid,
                                                                                                                          datasetSpec.datasetID,
                                                                                                                          datasetSpec.datasetName))
                                    for metadataName,metadaValue in iteritems(metaData):
                                        ddmIF.setDatasetMetadata(datasetSpec.datasetName,metadataName,metadaValue)
        # set lifetime to failed datasets
        if taskSpec.status in ['failed','broken','aborted']:
            trnLifeTime = 30*24*60*60
            ddmIF = self.ddmIF.getInterface(taskSpec.vo)
            # only log datasets
            metaData = {'lifetime':trnLifeTime}
            for datasetSpec in taskSpec.datasetSpecList:
                if datasetSpec.type in ['log']:
                    tmpLog.debug('set metadata={0} to failed datasetID={1}:Name={2}'.format(str(metaData),
                                                                                           datasetSpec.datasetID,
                                                                                           datasetSpec.datasetName))
                    for metadataName,metadaValue in iteritems(metaData):
                        ddmIF.setDatasetMetadata(datasetSpec.datasetName,metadataName,metadaValue)
        return self.SC_SUCCEEDED



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jediddm.DDMInterface import DDMInterface

from pandajedi.jediorder.JobBroker import JobBroker
from pandajedi.jediorder.JobSplitter import JobSplitter
from pandajedi.jediorder.JobGenerator import JobGeneratorThread
from pandajedi.jedicore.ThreadUtils import ThreadPool,ListWithLock
from pandajedi.jediorder.TaskSetupper import TaskSetupper
from pandajedi.jedicore import Interaction

import sys

logger = PandaLogger().getLogger('JobGenerator')
tmpLog = MsgWrapper(logger)


tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

siteMapper = tbIF.getSiteMapper()


ddmIF = DDMInterface()
ddmIF.setupInterface()

jediTaskID = int(sys.argv[1])
try:
    datasetID = [int(sys.argv[2])]
except Exception:
    datasetID = None

s,taskSpec = tbIF.getTaskWithID_JEDI(jediTaskID)

cloudName = taskSpec.cloud
vo = taskSpec.vo
prodSourceLabel = taskSpec.prodSourceLabel
queueID = taskSpec.workQueue_ID
gshare_name = taskSpec.gshare

workQueue = tbIF.getWorkQueueMap().getQueueWithIDGshare(queueID, gshare_name)

brokerageLockIDs = ListWithLock([])

threadPool = ThreadPool()

# get typical number of files
typicalNumFilesMap = tbIF.getTypicalNumInput_JEDI(vo,prodSourceLabel,workQueue,
                                                  useResultCache=600)

tmpListList = tbIF.getTasksToBeProcessed_JEDI(None,vo,workQueue,
                                              prodSourceLabel,
                                              cloudName,nFiles=10,simTasks=[jediTaskID],
                                              fullSimulation=True,
                                              typicalNumFilesMap=typicalNumFilesMap,
                                              simDatasets=datasetID,
                                              numNewTaskWithJumbo=5,
                                              ignore_lock=True)

taskSetupper = TaskSetupper(vo,prodSourceLabel)
taskSetupper.initializeMods(tbIF,ddmIF)

for dummyID,tmpList in tmpListList:
    task_common = {}
    for taskSpec,cloudName,inputChunk in tmpList:
        jobBroker = JobBroker(taskSpec.vo,taskSpec.prodSourceLabel)
        tmpStat = jobBroker.initializeMods(ddmIF.getInterface(vo),tbIF)
        jobBrokerCore = jobBroker.getImpl(taskSpec.vo, taskSpec.prodSourceLabel)
        jobBrokerCore.setTestMode()
        jobBrokerCore.set_task_common_dict(task_common)
        splitter = JobSplitter()
        gen = JobGeneratorThread(None,threadPool,tbIF,ddmIF,siteMapper,False,taskSetupper,None,
                                 None,'dummy',None,None,brokerageLockIDs, False)

        taskParamMap = None
        if taskSpec.useLimitedSites():
            tmpStat,taskParamMap = gen.readTaskParams(taskSpec,taskParamMap,tmpLog)
        jobBroker.setLockID(taskSpec.vo,taskSpec.prodSourceLabel,123,0)
        tmpStat,inputChunk = jobBroker.doBrokerage(taskSpec,cloudName,inputChunk,taskParamMap)
        brokerageLockID = jobBroker.getBaseLockID(taskSpec.vo,taskSpec.prodSourceLabel)
        if brokerageLockID is not None:
            brokerageLockIDs.append(brokerageLockID)
        for brokeragelockID in brokerageLockIDs:
            tbIF.unlockProcessWithPID_JEDI(taskSpec.vo,taskSpec.prodSourceLabel,workQueue.queue_id,
                                           brokeragelockID,True)
        tmpStat, subChunks, isSkipped = splitter.doSplit(taskSpec, inputChunk, siteMapper,
                                                         allow_chunk_size_limit=True)
        if tmpStat == Interaction.SC_SUCCEEDED and isSkipped:
            # run again without chunk size limit to generate jobs for skipped snippet
            tmpStat, tmpChunks, isSkipped = splitter.doSplit(taskSpec, inputChunk,
                                                             siteMapper,
                                                             allow_chunk_size_limit=False)
            if tmpStat == Interaction.SC_SUCCEEDED:
                subChunks += tmpChunks
        tmpStat,pandaJobs,datasetToRegister,oldPandaIDs,parallelOutMap,outDsMap = gen.doGenerate(taskSpec,cloudName,subChunks,inputChunk,tmpLog,True,
                                                                                                 splitter=splitter)



from pandajedi.jedicore.MsgWrapper import MsgWrapper

from .DDMClientBase import DDMClientBase

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])

# class to access to general DDM
class GenDDMClient(DDMClientBase):

    # constructor
    def __init__(self,con):
        # initialize base class
        DDMClientBase.__init__(self,con)

    # get dataset metadata
    def getDatasetMetaData(self, datasetName, ignore_missing=False):
        return self.SC_SUCCEEDED, {'state':'closed'}



import re
import time
import datetime
from pandajedi.jediconfig import jedi_config
from pandaserver.userinterface import Client
from pandacommon.pandalogger.PandaLogger import PandaLogger


class MsgWrapper:

    def __init__(self,logger,token=None,lineLimit=500,monToken=None):
        self.logger = logger
        # use timestamp as token if undefined
        if token is None:
            self.token = "<{0}>".format(datetime.datetime.utcnow().isoformat('/'))
        else:
            self.token = token
        # token for http logger
        if monToken is None:
            self.monToken = self.token
        else:
            self.monToken = monToken
        # remove <> for django
        try:
            self.monToken = re.sub('<(?P<name>[^>]+)>','\g<name>',self.monToken)
        except Exception:
            pass
        # message buffer
        self.msgBuffer = []
        self.bareMsg = []
        self.lineLimit = lineLimit


    def keepMsg(self,msg):
        # keep max message depth
        if len(self.msgBuffer) > self.lineLimit:
            self.msgBuffer.pop(0)
            self.bareMsg.pop(0)
        timeNow = datetime.datetime.utcnow()
        self.msgBuffer.append('{0} : {1}'.format(timeNow.isoformat(' '),msg))
        self.bareMsg.append(msg)


    def info(self,msg):
        msg = str(msg)
        self.logger.info(self.token + ' ' + msg)
        self.keepMsg(msg)


    def debug(self,msg):
        msg = str(msg)
        self.logger.debug(self.token + ' ' + msg)
        self.keepMsg(msg)


    def error(self,msg):
        msg = str(msg)
        self.logger.error(self.token + ' ' + msg)
        self.keepMsg(msg)


    def warning(self,msg):
        msg = str(msg)
        self.logger.warning(self.token + ' ' + msg)
        self.keepMsg(msg)


    def dumpToString(self):
        strMsg = ''
        for msg in self.msgBuffer:
            strMsg += msg
            strMsg += "\n"
        return strMsg


    def uploadLog(self,id):
        strMsg = self.dumpToString()
        s,o = Client.uploadLog(strMsg,id)
        if s != 0:
            return "failed to upload log with {0}.".format(s)
        if o.startswith('http'):
            return '<a href="{0}">log</a> : {1}.'.format(o, '. '.join(self.bareMsg[-2:]))
        return o


    # send message to logger
    def sendMsg(self,message,msgType,msgLevel='info',escapeChar=False):
        try:
            # get logger
            tmpPandaLogger = PandaLogger()
            # lock HTTP handler
            tmpPandaLogger.lock()
            tmpPandaLogger.setParams({'Type':msgType})
            # get logger
            tmpLogger = tmpPandaLogger.getHttpLogger(jedi_config.master.loggername)
            # escape special characters
            if escapeChar:
                message = message.replace('<','&lt;')
                message = message.replace('>','&gt;')
            # add message
            message = self.monToken + ' ' + message
            if msgLevel=='error':
                tmpLogger.error(message)
            elif msgLevel=='warning':
                tmpLogger.warning(message)
            elif msgLevel=='info':
                tmpLogger.info(message)
            else:
                tmpLogger.debug(message)                
            # release HTTP handler
            tmpPandaLogger.release()
        except Exception:
            pass



    # bulk send messages to logger
    def bulkSendMsg(self,msgType,msgLevel='info',loggerName=None):
        try:
            nChunk = 20
            if loggerName is None:
                loggerName = jedi_config.master.loggername
            for iMsg,message in enumerate(self.msgBuffer):
                # get logger
                tmpPandaLogger = PandaLogger()
                # lock HTTP handler
                tmpPandaLogger.lock()
                tmpPandaLogger.setParams({'Type':msgType})
                # get logger
                tmpLogger = tmpPandaLogger.getHttpLogger(loggerName)
                # add message
                message = self.monToken + ' ' + message
                if msgLevel=='error':
                    tmpLogger.error(message)
                elif msgLevel=='warning':
                    tmpLogger.warning(message)
                elif msgLevel=='info':
                    tmpLogger.info(message)
                else:
                    tmpLogger.debug(message)                
                # release HTTP handler
                tmpPandaLogger.release()
                if (iMsg+1) % nChunk == 0:
                    time.sleep(0.01)
        except Exception:
            pass



import re
from .TaskRefinerBase import TaskRefinerBase
from pandajedi.jedicore.JediTaskSpec import JediTaskSpec

# refiner for general purpose
class GenTaskRefiner (TaskRefinerBase):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        TaskRefinerBase.__init__(self,taskBufferIF,ddmIF)

    # extract common parameters
    def extractCommon(self, jediTaskID, taskParamMap, workQueueMapper, splitRule):
        if 'cloud' not in taskParamMap and 'workingGroup' in taskParamMap:
            taskParamMap['cloud'] = taskParamMap['workingGroup']
        if 'transPath' not in taskParamMap:
            taskParamMap['transPath'] = 'http://pandaserver-doma.cern.ch:25080/trf/user/runGen-00-00-02'
        # set sourceURL
        try:
            if 'sourceURL' in taskParamMap:
                for tmpItem in taskParamMap['jobParameters']:
                    if 'value' in tmpItem:
                        tmpItem['value'] = re.sub('\$\{SURL\}', taskParamMap['sourceURL'], tmpItem['value'])
        except Exception:
            pass
        # min ram count
        if 'ramCount' not in taskParamMap:
            taskParamMap['ramCount'] = 2000
            taskParamMap['ramUnit'] = 'MBPerCore'
        # push status changes
        if 'pushStatusChanges' not in taskParamMap:
            taskParamMap['pushStatusChanges'] = True
        # use cloud as VO
        taskParamMap['cloudAsVO'] = True
        # update task parameters
        self.updatedTaskParams = taskParamMap
        # call base method
        TaskRefinerBase.extractCommon(self, jediTaskID, taskParamMap, workQueueMapper, splitRule)

    # main
    def doRefine(self,jediTaskID,taskParamMap):
        # normal refine
        self.doBasicRefine(taskParamMap)
        # get DDM I/F to check
        if self.ddmIF.getInterface(self.taskSpec.vo, self.taskSpec.cloud):
            # use template dataset
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['instantiateTmpl'])
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['instantiateTmplSite'])
            for datasetSpec in self.outDatasetSpecList:
                datasetSpec.type = "tmpl_{0}".format(datasetSpec.type)
        return self.SC_SUCCEEDED



import json

from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin

from pandacommon.pandalogger import logger_utils


# logger
base_logger = logger_utils.setup_logger(__name__.split('.')[-1])


# Hyper-Parameter-Optimization message processing plugin
class HPOMsgProcPlugin(BaseMsgProcPlugin):

    def process(self, msg_obj, decoded_data=None):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, method_name='process')
        # start
        tmp_log.info('start')
        tmp_log.debug('sub_id={0} ; msg_id={1}'.format(msg_obj.sub_id, msg_obj.msg_id))
        # parse
        if decoded_data is None:
            # json decode
            try:
                msg_dict = json.loads(msg_obj.data)
            except Exception as e:
                err_str = 'failed to parse message json {2} , skipped. {0} : {1}'.format(e.__class__.__name__, e, msg_obj.data)
                tmp_log.error(err_str)
                raise
        else:
            msg_dict = decoded_data
        # sanity check
        try:
            msg_type = msg_dict['msg_type']
            jeditaskid = int(msg_dict['workload_id'])
            if msg_type == 'file_hyperparameteropt':
                target_list = msg_dict['files']
            elif msg_type == 'collection_hyperparameteropt':
                # to finish the task
                pass
            elif msg_type == 'work_hyperparameteropt':
                pass
            else:
                raise ValueError('invalid msg_type value: {0}'.format(msg_type))
        except Exception as e:
            err_str = 'failed to parse message object dict {2} , skipped. {0} : {1}'.format(e.__class__.__name__, e, msg_dict)
            tmp_log.error(err_str)
            raise
        # run
        if msg_type == 'file_hyperparameteropt':
            # insert HPO events
            try:
                # event ids from the targets
                event_id_list = []
                for target in target_list:
                    if target['status'] != 'New':
                        continue
                    model_id = None
                    try:
                        path = json.loads(target["path"])
                        if isinstance(path[0], (list, tuple)):
                            model_id = path[0][0]
                    except Exception:
                        pass
                    event_id_list.append((target['name'], model_id))
                if event_id_list:
                    n_events = len(event_id_list)
                    # insert events
                    res = self.tbIF.insertHpoEventAboutIdds_JEDI(   jedi_task_id=jeditaskid,
                                                                    event_id_list=event_id_list)
                    # check if ok
                    if res:
                        tmp_log.debug('jeditaskid={0}, inserted {1} events: {2}'.format(jeditaskid, n_events,
                                                                                        event_id_list))
                    else:
                        tmp_log.warning('jeditaskid={0}, failed to insert events: {1}'.format(jeditaskid,
                                                                                              event_id_list))
            except Exception as e:
                err_str = 'failed to parse message object, skipped. {0} : {1}'.format(e.__class__.__name__, e)
                tmp_log.error(err_str)
                raise
        elif msg_type == 'collection_hyperparameteropt':
            # finish the task
            try:

                # send finish command
                retVal, retStr = self.tbIF.sendCommandTaskPanda(jeditaskid,
                                                                'iDDS. HPO task finished',
                                                                True,
                                                                'finish',
                                                                comQualifier='soft')
                # check if ok
                if retVal:
                    tmp_log.debug('jeditaskid={0}, finished the task'.format(jeditaskid))
                else:
                    tmp_log.warning('jeditaskid={0}, failed finish the task: {1}'.format(jeditaskid, retStr))
            except Exception as e:
                err_str = 'failed to parse message object, skipped. {0} : {1}'.format(e.__class__.__name__, e)
                tmp_log.error(err_str)
                raise
        else:
            # do nothing
            tmp_log.debug('jeditaskid={jeditaskid}, msg_type={msg_type}, did nothing'.format(
                            jeditaskid=jeditaskid, msg_type=msg_type))
        # done
        tmp_log.info('done')



import re

from six import iteritems


"""
cache specification for JEDI

"""
class JediCacheSpec(object):
    # attributes
    attributes = (
            'main_key',
            'sub_key',
            'data',
            'last_update',
        )

    # constructor
    def __init__(self):
        # install attributes
        for attr in self.attributes:
            object.__setattr__(self, attr, None)

    # return map of values
    def valuesMap(self):
        ret = {}
        for attr in self.attributes:
            val = getattr(self, attr)
            ret[':%s' % attr] = val
        return ret

    # pack tuple into JediCacheSpec
    def pack(self, values):
        for i, attr in enumerate(self.attributes):
            val = values[i]
            object.__setattr__(self, attr, val)

    # return column names for INSERT
    @classmethod
    def columnNames(cls, prefix=None):
        ret = ''
        if prefix is None:
            ret = ','.join(cls.attributes)
        else:
            ret = ','.join([ '{0}.{1}'.format(prefix, attr) for attr in cls.attributes ])
        return ret

    # return expression of bind variables for INSERT
    @classmethod
    def bindValuesExpression(cls):
        values_str = ','.join([ ':{0}'.format(attr) for attr in cls.attributes ])
        ret = 'VALUES({0})'.format(values_str)
        return ret

    # return an expression of bind variables for UPDATE
    @classmethod
    def bindUpdateChangesExpression(cls):
        ret = ','.join([ '{0}=:{0}'.format(attr) for attr in cls.attributes ])
        return ret



from pandaserver import taskbuffer
import taskbuffer.DBProxyPool
from . import JediDBProxy

# use customized proxy
taskbuffer.DBProxyPool.DBProxy = JediDBProxy


class DBProxyPool(taskbuffer.DBProxyPool.DBProxyPool):
    # constructor
    def __init__(self, dbhost, dbpasswd, nConnection, useTimeout=False):
        taskbuffer.DBProxyPool.DBProxyPool.__init__(self, dbhost, dbpasswd,
                                                    nConnection, useTimeout)

    # get a DBProxyObj containing a proxy
    def get(self):
        proxy_obj = DBProxyObj(db_proxy_pool=self)
        return proxy_obj


# object of context manager for db proxy
class DBProxyObj(object):
    # constructor
    def __init__(self, db_proxy_pool):
        self.proxy_pool = db_proxy_pool
        self.proxy = None

    # get proxy
    def __enter__(self):
        self.proxy = self.proxy_pool.getProxy()
        return self.proxy

    # release proxy
    def __exit__(self, type, value, traceback):
        self.proxy_pool.putProxy(self.proxy)
        self.proxy = None



import sys
from .MsgWrapper import MsgWrapper

_factoryModuleName = __name__.split('.')[-1]

# base class for factory
class FactoryBase:

    # constructor
    def __init__(self,vos,sourceLabels,logger,modConfig):
        if isinstance(vos,list):
            self.vos = vos
        else:
            try:
                self.vos = vos.split('|')
            except Exception:
                self.vos = [vos]
        if isinstance(sourceLabels,list):
            self.sourceLabels = sourceLabels
        else:
            try:
                self.sourceLabels = sourceLabels.split('|')
            except Exception:
                self.sourceLabels = [sourceLabels]
        self.modConfig = modConfig
        self.logger = MsgWrapper(logger,_factoryModuleName)
        self.implMap = {}
        self.className = None
        self.classMap = {}


    # initialize all modules
    def initializeMods(self,*args):
        # parse config
        for configStr in self.modConfig.split(','):
            configStr = configStr.strip()
            items = configStr.split(':')
            # check format
            try:
                vos          = items[0].split('|')
                sourceLabels = items[1].split('|')
                moduleName   = items[2]
                className    = items[3]
                try:
                    subTypes = items[4].split('|')
                except Exception:
                    subTypes = ['any']
            except Exception:
                self.logger.error('wrong config definition : {0}'.format(configStr))
                continue
            # loop over all VOs
            for vo in vos:
                # loop over all labels
                for sourceLabel in sourceLabels:
                    # check vo and sourceLabel if specified
                    if vo not in ['','any'] and \
                            vo not in  self.vos and \
                            None not in  self.vos and \
                            'any' not in  self.vos:
                        continue
                    if sourceLabel not in ['','any'] and \
                            sourceLabel not in  self.sourceLabels and \
                            None not in  self.sourceLabels and \
                            'any' not in  self.sourceLabels:
                        continue
                    # loop over all sub types
                    for subType in subTypes:
                        # import
                        try:
                            # import module
                            self.logger.info("vo={0} label={1} subtype={2}".format(vo,sourceLabel,subType))
                            self.logger.info("importing {0}".format(moduleName))
                            mod = __import__(moduleName)
                            for subModuleName in moduleName.split('.')[1:]:
                                mod = getattr(mod,subModuleName)
                            # get class
                            self.logger.info("getting class {0}".format(className))
                            cls = getattr(mod,className)
                            # instantiate
                            self.logger.info("instantiating")
                            impl = cls(*args)
                            # set vo
                            impl.vo = vo
                            impl.prodSourceLabel = sourceLabel
                            # append
                            if vo not in self.implMap:
                                self.implMap[vo] = {}
                                self.classMap[vo] = {}
                            if sourceLabel not in self.implMap[vo]:
                                self.implMap[vo][sourceLabel] = {}
                                self.classMap[vo][sourceLabel] = {}
                            self.implMap[vo][sourceLabel][subType] = impl
                            self.classMap[vo][sourceLabel][subType] = cls
                            self.logger.info("{0} is ready for {1}:{2}:{3}".format(cls,vo,sourceLabel,subType))
                        except Exception:
                            errtype,errvalue = sys.exc_info()[:2]
                            self.logger.error(
                                'failed to import {mn}.{cn} for vo={vo} label={lb} subtype={st} due to {et} {ev}'.format(et=errtype.__name__,
                                                                                                                         ev=errvalue,
                                                                                                                         st=subType,
                                                                                                                         vo=vo,
                                                                                                                         lb=sourceLabel,
                                                                                                                         cn=className,
                                                                                                                         mn=moduleName)
                                )
                            raise ImportError('failed to import {0}.{1}'.format(moduleName,className))
        # return
        return True


    # get implementation for vo and sourceLabel. Only work with initializeMods()
    def getImpl(self,vo,sourceLabel,subType='any',doRefresh=True):
        # check VO
        if vo in self.implMap:
            # match VO
            voImplMap = self.implMap[vo]
        elif 'any' in self.implMap:
            # catch all
            voImplMap = self.implMap['any']
        else:
            return None
        # check sourceLabel
        if sourceLabel in voImplMap:
            # match sourceLabel
            srcImplMap = voImplMap[sourceLabel]
        elif 'any' in voImplMap:
            # catch all
            srcImplMap = voImplMap['any']
        else:
            return None
        # check subType
        if subType in srcImplMap:
            # match subType
            tmpImpl = srcImplMap[subType]
            if doRefresh:
                tmpImpl.refresh()
            return tmpImpl
        elif 'any' in srcImplMap:
            # catch all
            tmpImpl = srcImplMap['any']
            if doRefresh:
                tmpImpl.refresh()
            return tmpImpl
        else:
            return None


    # instantiate implementation for vo and sourceLabel. Only work with initializeMods()
    def instantiateImpl(self,vo,sourceLabel,subType,*args):
        # check VO
        if vo in self.classMap:
            # match VO
            voImplMap = self.classMap[vo]
        elif 'any' in self.classMap:
            # catch all
            voImplMap = self.classMap['any']
        else:
            return None
        # check sourceLabel
        if sourceLabel in voImplMap:
            # match sourceLabel
            srcImplMap = voImplMap[sourceLabel]
        elif 'any' in voImplMap:
            # catch all
            srcImplMap = voImplMap['any']
        else:
            return None
        # check subType
        if subType in srcImplMap:
            # match subType
            impl = srcImplMap[subType](*args)
            impl.vo = vo
            impl.prodSourceLabel= sourceLabel
            return impl
        elif 'any' in srcImplMap:
            # catch all
            impl = srcImplMap['any'](*args)
            impl.vo = vo
            impl.prodSourceLabel= sourceLabel
            return impl
        else:
            return None



    # get class name of impl
    def getClassName(self,vo=None,sourceLabel=None):
        impl = self.getImpl(vo,sourceLabel,doRefresh=False)
        if impl is None:
            return None
        return impl.__class__.__name__



import re
import sys

from pandajedi.jedicore import Interaction
from .PostProcessorBase import PostProcessorBase



# post processor for general purpose
class GenPostProcessor (PostProcessorBase):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        PostProcessorBase.__init__(self,taskBufferIF,ddmIF)


    # main
    def doPostProcess(self,taskSpec,tmpLog):
        try:
            self.doBasicPostProcess(taskSpec,tmpLog)
        except Exception:
            errtype,errvalue = sys.exc_info()[:2]
            tmpLog.error('doBasicPostProcess failed with {0}:{1}'.format(errtype.__name__,errvalue))
            return self.SC_FATAL
        return self.SC_SUCCEEDED
            
    



from pandajedi.jedicore import Interaction
from pandajedi.jedicore.InputChunk import InputChunk
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore import JediCoreUtils

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# class to split job
class JobSplitter:

    # constructor
    def __init__(self):
        self.sizeGradientsPerInSizeForMerge = 1.2
        self.interceptsMerginForMerge = 500 * 1024 * 1024



    # split
    def doSplit(self, taskSpec, inputChunk, siteMapper, allow_chunk_size_limit=False):
        # return for failure
        retFatal    = self.SC_FATAL,[]
        retTmpError = self.SC_FAILED,[]
        # make logger
        tmpLog = MsgWrapper(logger,'< jediTaskID={0} datasetID={1} >'.format(taskSpec.jediTaskID,inputChunk.masterIndexName))
        tmpLog.debug('start chunk_size_limit={}'.format(allow_chunk_size_limit))
        if not inputChunk.isMerging:
            # set maxNumFiles using taskSpec if specified
            maxNumFiles = taskSpec.getMaxNumFilesPerJob()
            # set fsize gradients using taskSpec
            sizeGradients  = taskSpec.getOutDiskSize()
            # set fsize intercepts using taskSpec
            sizeIntercepts = taskSpec.getWorkDiskSize()
            # walltime
            if not taskSpec.useHS06():
                walltimeGradient = taskSpec.walltime
            else:
                walltimeGradient = taskSpec.getCpuTime()
            # number of events per job if defined
            nEventsPerJob = taskSpec.getNumEventsPerJob()
            # number of files per job if defined
            if not taskSpec.dynamicNumEvents():
                nFilesPerJob = taskSpec.getNumFilesPerJob()
            else:
                nFilesPerJob = None
            if nFilesPerJob is None and nEventsPerJob is None and inputChunk.useScout() \
                    and not taskSpec.useLoadXML() and not taskSpec.respectSplitRule():
                nFilesPerJob = 1
            # grouping with boundaryID
            useBoundary = taskSpec.useGroupWithBoundaryID()
            # fsize intercepts per input size
            sizeGradientsPerInSize = None
            # max primay output size
            maxOutSize = None
            # max size per job
            maxSizePerJob = taskSpec.getMaxSizePerJob()
            if maxSizePerJob is not None:
                maxSizePerJob += InputChunk.defaultOutputSize
            # dynamic number of events
            dynNumEvents = taskSpec.dynamicNumEvents()
            # max number of event ranges
            maxNumEventRanges = None
            # multiplicity of jobs
            if taskSpec.useJobCloning():
                multiplicity = 1
            else:
                multiplicity = taskSpec.getNumEventServiceConsumer()
            # split with fields
            if taskSpec.getFieldNumToLFN() is not None and taskSpec.useFileAsSourceLFN():
                splitByFields = taskSpec.getFieldNumToLFN()
            else:
                splitByFields = None
        else:
            # set parameters for merging
            maxNumFiles = taskSpec.getMaxNumFilesPerMergeJob()
            sizeGradients = 0
            walltimeGradient = 0
            nFilesPerJob = taskSpec.getNumFilesPerMergeJob()
            nEventsPerJob = taskSpec.getNumEventsPerMergeJob()
            maxSizePerJob = None
            useBoundary = {'inSplit':3}
            dynNumEvents = False
            maxNumEventRanges = None
            multiplicity = None
            # gradients per input size is 1 + margin
            sizeGradientsPerInSize = self.sizeGradientsPerInSizeForMerge
            # intercepts for libDS
            sizeIntercepts = taskSpec.getWorkDiskSize()
            # mergein of 500MB
            interceptsMergin = self.interceptsMerginForMerge
            if sizeIntercepts < interceptsMergin:
                sizeIntercepts = interceptsMergin
            maxOutSize = taskSpec.getMaxSizePerMergeJob()
            if maxOutSize is None:
                # max output size is 5GB for merging by default
                maxOutSize = 5 * 1024 * 1024 * 1024
            # split with fields
            if taskSpec.getFieldNumToLFN() is not None and taskSpec.useFileAsSourceLFN():
                splitByFields = list(range(4+1, 4+1+len(taskSpec.getFieldNumToLFN())))
            else:
                splitByFields = None
        # LB
        respectLB = taskSpec.respectLumiblock()
        # dump
        tmpLog.debug('maxNumFiles={0} sizeGradients={1} sizeIntercepts={2} useBoundary={3}'.format(maxNumFiles,
                                                                                                   sizeGradients,
                                                                                                   sizeIntercepts,
                                                                                                   useBoundary))
        tmpLog.debug('walltimeGradient={0} nFilesPerJob={1} nEventsPerJob={2}'.format(walltimeGradient,
                                                                                        nFilesPerJob,
                                                                                        nEventsPerJob))
        tmpLog.debug('useScout={} isMerging={}'.format(inputChunk.useScout(), inputChunk.isMerging))
        tmpLog.debug('sizeGradientsPerInSize={0} maxOutSize={1} respectLB={2} dynNumEvents={3}'.format(sizeGradientsPerInSize,
                                                                                                       maxOutSize,
                                                                                                       respectLB,
                                                                                                       dynNumEvents))
        tmpLog.debug('multiplicity={0} splitByFields={1} nFiles={2}'.format(multiplicity,str(splitByFields),
                                                                            inputChunk.getNumFilesInMaster()))
        # split
        returnList = []
        subChunks  = []
        iSubChunks = 0
        if inputChunk.useScout() and not inputChunk.isMerging:
            default_nSubChunks = 2
        elif taskSpec.is_hpo_workflow():
            default_nSubChunks = 2
        else:
            default_nSubChunks = 25
        subChunk   = None
        nSubChunks = default_nSubChunks
        strict_chunkSize = False
        while True:
            # change site
            if iSubChunks % nSubChunks == 0 or subChunk == []:
                # append to return map
                if subChunks != []:
                    # get site names for parallel execution
                    if taskSpec.getNumSitesPerJob() > 1 and not inputChunk.isMerging and inputChunk.useJumbo != 'fake':
                        siteName = inputChunk.getParallelSites(taskSpec.getNumSitesPerJob(),
                                                               nSubChunks,[siteName])
                    returnList.append({'siteName':siteName,
                                       'subChunks':subChunks,
                                       'siteCandidate':siteCandidate,
                                       })
                    try:
                        gshare = taskSpec.gshare.replace(' ', '_')
                    except Exception:
                        gshare = None
                    tmpLog.info('split to nJobs=%s at site=%s gshare=%s' % (len(subChunks), siteName,
                                                                                       gshare))
                    # checkpoint
                    inputChunk.checkpoint_file_usage()
                    # reset
                    subChunks = []
                # skip PQs with chunk size limit
                ngList = []
                if not allow_chunk_size_limit:
                    for siteName in inputChunk.get_candidate_names():
                        siteSpec = siteMapper.getSite(siteName)
                        if siteSpec.get_job_chunk_size() is not None:
                            ngList.append(siteName)
                # new candidate
                siteCandidate, getCandidateMsg = inputChunk.getOneSiteCandidate(nSubChunks, ngSites=ngList,
                                                                                get_msg=True)
                if siteCandidate is None:
                    tmpLog.debug('no candidate: {0}'.format(getCandidateMsg))
                    break
                siteName = siteCandidate.siteName
                siteSpec = siteMapper.getSite(siteName)
                # set chunk size
                nSubChunks = siteSpec.get_job_chunk_size()
                if nSubChunks is None:
                    nSubChunks = default_nSubChunks
                    strict_chunkSize = False
                else:
                    strict_chunkSize = True
                # directIO
                if not JediCoreUtils.use_direct_io_for_job(taskSpec, siteSpec, inputChunk):
                    useDirectIO = False
                else:
                    useDirectIO = True
                # get maxSize if it is set in taskSpec
                maxSize = maxSizePerJob
                if maxSize is None:
                    # use maxwdir as the default maxSize
                    if not useDirectIO:
                        maxSize = siteCandidate.get_overridden_attribute('maxwdir')
                        if maxSize is None:
                            maxSize = siteSpec.maxwdir
                        if maxSize:
                            maxSize *= 1024 * 1024
                    elif nEventsPerJob is not None or nFilesPerJob is not None:
                        maxSize = None
                    else:
                        maxSize = siteCandidate.get_overridden_attribute('maxwdir')
                        if maxSize is None:
                            maxSize = siteSpec.maxwdir
                        maxSize = max(50000, maxSize) * 1024 * 1024
                else:
                    # add offset
                    maxSize += sizeIntercepts
                # max disk size
                maxDiskSize = siteCandidate.get_overridden_attribute('maxwdir')
                if maxDiskSize is None:
                    maxDiskSize = siteSpec.maxwdir
                if maxDiskSize:
                    maxDiskSize *= 1024 * 1024
                # max walltime
                maxWalltime = None
                if not inputChunk.isMerging:
                    maxWalltime = taskSpec.getMaxWalltime()
                if maxWalltime is None:
                    maxWalltime = siteSpec.maxtime
                # core count
                if siteSpec.coreCount:
                    coreCount = siteSpec.coreCount
                else:
                    coreCount = 1
                # core power
                corePower = siteSpec.corepower
                # max num of event ranges for dynNumEvents
                if dynNumEvents:
                    maxNumEventRanges = int(siteSpec.get_n_sim_events() // taskSpec.get_min_granularity())
                    if maxNumEventRanges == 0:
                        maxNumEventRanges = 1
                tmpLog.debug('chosen {0} : {1} : nQueue={2} nRunCap={3}'.format(siteName, getCandidateMsg,
                                                                                siteCandidate.nQueuedJobs,
                                                                                siteCandidate.nRunningJobsCap))
                tmpLog.debug('new weight {0}'.format(siteCandidate.weight))
                tmpLog.debug('maxSize={0} maxWalltime={1} coreCount={2} corePower={3} maxNumEventRanges={4} maxDisk={5}'.format(maxSize,maxWalltime,
                                                                                                                                coreCount,corePower,
                                                                                                                                maxNumEventRanges,
                                                                                                                                maxDiskSize))
                tmpLog.debug('useDirectIO={0} label={1}'.format(useDirectIO, taskSpec.prodSourceLabel))
            # get sub chunk
            subChunk = inputChunk.getSubChunk(siteName,maxSize=maxSize,
                                              maxNumFiles=maxNumFiles,
                                              sizeGradients=sizeGradients,
                                              sizeIntercepts=sizeIntercepts,
                                              nFilesPerJob=nFilesPerJob,
                                              walltimeGradient=walltimeGradient,
                                              maxWalltime=maxWalltime,
                                              nEventsPerJob=nEventsPerJob,
                                              useBoundary=useBoundary,
                                              sizeGradientsPerInSize=sizeGradientsPerInSize,
                                              maxOutSize=maxOutSize,
                                              coreCount=coreCount,
                                              respectLB=respectLB,
                                              corePower=corePower,
                                              dynNumEvents=dynNumEvents,
                                              maxNumEventRanges=maxNumEventRanges,
                                              multiplicity=multiplicity,
                                              splitByFields=splitByFields,
                                              tmpLog=tmpLog,
                                              useDirectIO=useDirectIO,
                                              maxDiskSize=maxDiskSize,
                                              enableLog=True)
            if subChunk is None:
                break
            if subChunk != []:
                # append
                subChunks.append(subChunk)
            iSubChunks += 1
        # append to return map if remain
        isSkipped = False
        if subChunks != []:
            # skip if chunk size is not enough
            if allow_chunk_size_limit and strict_chunkSize and len(subChunks) < nSubChunks:
                tmpLog.debug('skip splitting since chunk size {} is less than chunk size limit {} at {}'.format(
                    len(subChunks), nSubChunks, siteName))
                inputChunk.rollback_file_usage()
                isSkipped = True
            else:
                # get site names for parallel execution
                if taskSpec.getNumSitesPerJob() > 1 and not inputChunk.isMerging:
                    siteName = inputChunk.getParallelSites(taskSpec.getNumSitesPerJob(),
                                                           nSubChunks,[siteName])
                returnList.append({'siteName':siteName,
                                   'subChunks':subChunks,
                                   'siteCandidate':siteCandidate,
                                   })
                try:
                    gshare = taskSpec.gshare.replace(' ', '_')
                except Exception:
                    gshare = None
                tmpLog.info('split to nJobs=%s at site=%s gshare=%s' % (len(subChunks), siteName,
                                                                                   gshare))
        # return
        tmpLog.debug('done')
        return self.SC_SUCCEEDED, returnList, isSkipped



Interaction.installSC(JobSplitter)



import copy
from pandajedi.jedicore import JediException
from pandajedi.jedirefine import RefinerUtils
try:
    from idds.client.client import Client as iDDS_Client
    import idds.common.constants
    import idds.common.utils
except ImportError:
    pass


# send notification to external system for additional post-processing
def send_notification(taskBufferIF, ddmIF, taskSpec, tmpLog):
    # send notification to external system
    try:
        taskParam = taskBufferIF.getTaskParamsWithID_JEDI(taskSpec.jediTaskID)
        taskParamMap = RefinerUtils.decodeJSON(taskParam)
    except Exception as e:
        errStr = 'task param conversion from json failed with {0}'.format(str(e))
        raise JediException.ExternalTempError(errStr)
    if 'outputPostProcessing' in taskParamMap and \
            'system' in taskParamMap['outputPostProcessing']:
        if taskParamMap['outputPostProcessing']['system'] == 'idds':
            try:
                c = iDDS_Client(idds.common.utils.get_rest_host())
                if taskParamMap['outputPostProcessing']['type'] == 'active_learning':
                    for datasetSpec in taskSpec.datasetSpecList:
                        if datasetSpec.type != 'output':
                            continue
                        data = copy.copy(taskParamMap['outputPostProcessing']['data'])
                        tmp_scope, tmp_name = ddmIF.extract_scope(datasetSpec.datasetName)
                        data['workload_id'] = taskSpec.jediTaskID
                        req = {
                            'scope': tmp_scope,
                            'name': tmp_name,
                            'requester': 'panda',
                            'request_type': idds.common.constants.RequestType.ActiveLearning,
                            'transform_tag': idds.common.constants.RequestType.ActiveLearning.value,
                            'status': idds.common.constants.RequestStatus.New,
                            'priority': 0,
                            'lifetime': 30,
                            'request_metadata': data,
                        }
                        tmpLog.debug('req {0}'.format(str(req)))
                        ret = c.add_request(**req)
                        tmpLog.debug('got requestID={0}'.format(str(ret)))
            except Exception as e:
                errStr = 'iDDS failed with {0}'.format(str(e))
                raise JediException.ExternalTempError(errStr)


from pandajedi.jedicore import Interaction

# throttle level
THR_LEVEL5 = 5

# base class for job throttle
class JobThrottlerBase (object):

    def __init__(self,taskBufferIF):
        self.taskBufferIF = taskBufferIF
        # returns
        self.retTmpError    = self.SC_FAILED,True
        self.retThrottled   = self.SC_SUCCEEDED,True
        self.retUnThrottled = self.SC_SUCCEEDED,False
        self.retMergeUnThr  = self.SC_SUCCEEDED,THR_LEVEL5
        # limit
        self.refresh()
        self.msgType      = 'jobthrottler'


    # refresh
    def refresh(self):
        self.maxNumJobs  = None
        self.minPriority = None
        self.underNqLimit = False
        self.siteMapper = self.taskBufferIF.getSiteMapper()

        
    # set maximum number of jobs to be submitted    
    def setMaxNumJobs(self,maxNumJobs):
        self.maxNumJobs = maxNumJobs


    # set min priority of jobs to be submitted
    def setMinPriority(self,minPriority):
        self.minPriority = minPriority


    # check throttle level
    def mergeThrottled(self,thrLevel):
        # un-leveled flag
        if thrLevel in [True,False]:
            return thrLevel
        return thrLevel > THR_LEVEL5


    # check if lack of jobs
    def lackOfJobs(self):
        return self.underNqLimit


    # not enough jobs are queued
    def notEnoughJobsQueued(self):
        self.underNqLimit = True

        


Interaction.installSC(JobThrottlerBase)



import re
import sys
import copy

from .import RefinerUtils
from pandajedi.jedicore import Interaction
from .TaskRefinerBase import TaskRefinerBase
from pandajedi.jedicore.JediTaskSpec import JediTaskSpec
from pandajedi.jedicore.JediDatasetSpec import JediDatasetSpec

# refiner for lost file recovery
class FileRecovery (TaskRefinerBase):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        TaskRefinerBase.__init__(self,taskBufferIF,ddmIF)


    # extract common parameters
    def extractCommon(self,jediTaskID,taskParamMap,workQueueMapper):
        return self.SC_SUCCEEDED


    # check matching of dataset names
    def checkDatasetNameMatching(self,datasetName,patternList):
        # name list is not given
        if patternList is None:
            return False
        # loop over all names
        for namePattern in patternList:
            if re.search('^'+namePattern+'$',datasetName) is not None:
                return True
        return False


    # main
    def doRefine(self,jediTaskID,taskParamMap):
        try:
            # make logger
            tmpLog = self.tmpLog
            tmpLog.debug('start jediTaskID={0}'.format(jediTaskID))
            # old dataset name
            oldDatasetName = taskParamMap['oldDatasetName']
            # accompany datasets
            if 'oldAccompanyDatasetNames' in taskParamMap:
                oldAccDatasetNames = taskParamMap['oldAccompanyDatasetNames']
            else:
                oldAccDatasetNames = None
            # use first file to get task and dataset info
            lostFileName = taskParamMap['lostFiles'][0]
            # get ole jediTaskID and datasetIDs
            tmpStat,oldIDs = self.taskBufferIF.getIDsWithFileDataset_JEDI(oldDatasetName,lostFileName,'output')
            if tmpStat is not True or oldIDs is None:
                tmpLog.error('failed to get jediTaskID and DatasetID for {0}:{1}'.format(oldDatasetName,
                                                                                         lostFileName))
                return self.SC_FAILED
            # get task
            oldJediTaskID = oldIDs['jediTaskID']
            oldDatasetID  = oldIDs['datasetID']
            tmpStat,oldTaskSpec = self.taskBufferIF.getTaskWithID_JEDI(oldJediTaskID,True)
            if tmpStat is not True:
                tmpLog.error('failed to get TaskSpec for old jediTaskId={0}'.format(oldJediTaskID))
                return self.SC_FAILED
            # make task spec
            taskSpec = JediTaskSpec()
            taskSpec.copyAttributes(oldTaskSpec)
            # reset attributes
            taskSpec.jediTaskID   = jediTaskID
            taskSpec.taskType     = taskParamMap['taskType']
            taskSpec.taskPriority = taskParamMap['taskPriority']
            self.taskSpec = taskSpec
            # get datasets
            tmpStat,datasetSpecList = self.taskBufferIF.getDatasetsWithJediTaskID_JEDI(oldJediTaskID)
            if tmpStat is not True:
                tmpLog.error('failed to get datasetSpecs')
                return self.SC_FAILED
            # loop over all datasets
            provenanceID = None
            dummyStreams = []
            outDatasetSpec = None
            datasetNameSpecMap = {}
            for datasetSpec in datasetSpecList:
                # for output datasets
                if datasetSpec.type not in JediDatasetSpec.getInputTypes():
                    # collect output with the same provenanceID
                    if provenanceID is not None and datasetSpec.provenanceID != provenanceID:
                        continue
                    # set provenanceID if undefined
                    if provenanceID is None and datasetSpec.provenanceID is not None:
                        provenanceID = datasetSpec.provenanceID
                    # collect dummy streams
                    if datasetSpec.type != 'log' and (datasetSpec.datasetID != oldDatasetID and \
                                                          not self.checkDatasetNameMatching(datasetSpec.datasetName,oldAccDatasetNames)):
                        if datasetSpec.streamName not in dummyStreams:
                            dummyStreams.append(datasetSpec.streamName)
                        continue
                # reset attributes
                datasetSpec.status = 'defined'
                datasetSpec.datasetID  = None
                datasetSpec.jediTaskID = jediTaskID
                datasetSpec.nFiles = 0
                datasetSpec.nFilesUsed = 0
                datasetSpec.nFilesToBeUsed = 0
                datasetSpec.nFilesFinished = 0
                datasetSpec.nFilesFailed   = 0
                datasetSpec.nFilesOnHold   = 0
                # remove nosplit and repeat since even the same file is made for each bounaryID
                datasetSpec.remNoSplit()
                datasetSpec.remRepeat()
                # append to map
                datasetNameSpecMap[datasetSpec.datasetName] = datasetSpec
                # set master and secondary for input
                if datasetSpec.type in JediDatasetSpec.getInputTypes():
                    if datasetSpec.isMaster():
                        # master
                        self.inMasterDatasetSpec = datasetSpec
                    else:
                        # secondary
                        self.inSecDatasetSpecList.append(datasetSpec)
                elif datasetSpec.type == 'log':
                    # set new attributes
                    tmpItem = taskParamMap['log']
                    datasetSpec.datasetName = tmpItem['dataset']
                    if 'container' in tmpItem:
                        datasetSpec.containerName = tmpItem['container']
                    if 'token' in tmpItem:
                        datasetSpec.storageToken = tmpItem['token']
                    if 'destination' in tmpItem:
                        datasetSpec.destination = tmpItem['destination']
                    # extract output filename template and change the value field
                    outFileTemplate,tmpItem['value'] = RefinerUtils.extractReplaceOutFileTemplate(tmpItem['value'],
                                                                                                  datasetSpec.streamName)
                    # make output template
                    if outFileTemplate is not None:
                        if 'offset' in tmpItem:
                            offsetVal = 1 + tmpItem['offset']
                        else:
                            offsetVal = 1
                        outTemplateMap = {'jediTaskID' : self.taskSpec.jediTaskID,
                                          'serialNr' : offsetVal,
                                          'streamName' : datasetSpec.streamName,
                                          'filenameTemplate' : outFileTemplate,
                                          'outtype' : datasetSpec.type,
                                          }
                        self.outputTemplateMap[datasetSpec.outputMapKey()] = [outTemplateMap]
                    # append
                    self.outDatasetSpecList.append(datasetSpec)
                else:
                    # output dataset to make copies later
                    outDatasetSpec = datasetSpec
            # replace redundant output streams with dummy files
            for dummyStream in dummyStreams:
                self.taskSpec.jobParamsTemplate = self.taskSpec.jobParamsTemplate.replace('${'+dummyStream+'}',
                                                                                          dummyStream.lower()+'.tmp')
            self.setJobParamsTemplate(self.taskSpec.jobParamsTemplate)
            # loop over all lost files
            datasetIDSpecMap = {}
            for lostFileName in taskParamMap['lostFiles']:
                # get FileID
                tmpStat,tmpIDs = self.taskBufferIF.getIDsWithFileDataset_JEDI(oldDatasetName,lostFileName,'output')
                if tmpStat is not True or tmpIDs is None:
                    tmpLog.error('failed to get FileID for {0}:{1}'.format(oldDatasetName,
                                                                           lostFileName))
                    return self.SC_FAILED
                # get PandaID
                tmpStat,pandaID = self.taskBufferIF.getPandaIDWithFileID_JEDI(tmpIDs['jediTaskID'],
                                                                              tmpIDs['datasetID'],
                                                                              tmpIDs['fileID'])
                if tmpStat is not True or pandaID is None:
                    tmpLog.error('failed to get PandaID for {0}'.format(str(tmpIDs)))
                    return self.SC_FAILED
                # get files
                tmpStat,fileSpecList = self.taskBufferIF.getFilesWithPandaID_JEDI(pandaID)
                if tmpStat is not True or fileSpecList == []:
                    tmpLog.error('failed to get files for PandaID={0}'.format(pandaID))
                    return self.SC_FAILED
                # append
                for fileSpec in fileSpecList:
                    # only input types
                    if fileSpec.type not in JediDatasetSpec.getInputTypes():
                        continue
                    # get original datasetSpec
                    if fileSpec.datasetID not in datasetIDSpecMap:
                        tmpStat,tmpDatasetSpec = self.taskBufferIF.getDatasetWithID_JEDI(fileSpec.jediTaskID,fileSpec.datasetID)
                        if tmpStat is not True or tmpDatasetSpec is None:
                            tmpLog.error('failed to get dataset for jediTaskID={0} datasetID={1}'.format(fileSpec.jediTaskID,
                                                                                                         fileSpec.datasetID))
                            return self.SC_FAILED
                        datasetIDSpecMap[fileSpec.datasetID] = tmpDatasetSpec
                    origDatasetSpec = datasetIDSpecMap[fileSpec.datasetID]
                    if origDatasetSpec.datasetName not in datasetNameSpecMap:
                        tmpLog.error('datasetName={0} is missing in new datasets'.format(origDatasetSpec.datasetName))
                        return self.SC_FAILED
                    # not target or accompany datasets
                    if origDatasetSpec.datasetID != oldDatasetID and \
                            not self.checkDatasetNameMatching(origDatasetSpec.datasetName,oldAccDatasetNames):
                        continue
                    newDatasetSpec = datasetNameSpecMap[origDatasetSpec.datasetName]
                    # set new attributes
                    fileSpec.fileID = None
                    fileSpec.datasetID = None
                    fileSpec.jediTaskID = None
                    fileSpec.boundaryID = pandaID
                    fileSpec.keepTrack = 1
                    fileSpec.attemptNr = 1
                    fileSpec.status = 'ready'
                    # append
                    newDatasetSpec.addFile(fileSpec)
                # make one output dataset per file
                datasetSpec = copy.copy(outDatasetSpec)
                # set new attributes
                tmpItem = taskParamMap['output']
                datasetSpec.datasetName = tmpItem['dataset']
                if 'container' in tmpItem:
                    datasetSpec.containerName = tmpItem['container']
                if 'token' in tmpItem:
                    datasetSpec.storageToken = tmpItem['token']
                if 'destination' in tmpItem:
                    datasetSpec.destination = tmpItem['destination']
                # use PandaID of original job as provenanceID
                datasetSpec.provenanceID = pandaID
                # append
                self.outDatasetSpecList.append(datasetSpec)
                # extract attempt number from original filename
                tmpMatch = re.search('\.(\d+)$',lostFileName)
                if tmpMatch is None:
                    offsetVal = 1
                else:
                    offsetVal = 1 + int(tmpMatch.group(1))
                # filename without attempt number
                baseFileName = re.sub('\.(\d+)$','',lostFileName)
                # make output template
                outTemplateMap = {'jediTaskID' : self.taskSpec.jediTaskID,
                                  'serialNr' : offsetVal,
                                  'streamName' : datasetSpec.streamName,
                                  'filenameTemplate' : baseFileName + '.${SN:d}',
                                  'outtype' : datasetSpec.type,
                                  }
                self.outputTemplateMap[datasetSpec.outputMapKey()] = [outTemplateMap]
            # append datasets to task parameters
            for datasetSpec in datasetNameSpecMap.values():
                if datasetSpec.Files == []:
                    continue
                fileList = []
                for fileSpec in datasetSpec.Files:
                    fileList.append({'lfn':fileSpec.lfn,
                                     'firstEvent':fileSpec.firstEvent,
                                     'startEvent':fileSpec.startEvent,
                                     'endEvent':fileSpec.endEvent,
                                     'keepTrack':fileSpec.keepTrack,
                                     'boundaryID':fileSpec.boundaryID,
                                     })
                taskParamMap = RefinerUtils.appendDataset(taskParamMap,datasetSpec,fileList)
                self.updatedTaskParams = taskParamMap
            # grouping with boundaryID
            self.setSplitRule(None,4,JediTaskSpec.splitRuleToken['groupBoundaryID'])
        except Exception:
            errtype,errvalue = sys.exc_info()[:2]
            tmpLog.error('doRefine failed with {0}:{1}'.format(errtype.__name__,errvalue))
            return self.SC_FAILED
        tmpLog.debug('done')
        return self.SC_SUCCEEDED



from pandajedi.jedicore.FactoryBase import FactoryBase
from pandajedi.jediconfig import jedi_config

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# factory class for task setup
class TaskSetupper (FactoryBase):

    # constructor
    def __init__(self,vo,sourceLabel):
        FactoryBase.__init__(self,vo,sourceLabel,logger,
                             jedi_config.tasksetup.modConfig)


    # main
    def doSetup(self,taskSpec,datasetToRegister,pandaJobs):
        return self.getImpl(taskSpec.vo,taskSpec.prodSourceLabel).doSetup(taskSpec,datasetToRegister,pandaJobs)



from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore import Interaction
from .TaskSetupperBase import TaskSetupperBase


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# task setup for general purpose
class GenTaskSetupper (TaskSetupperBase):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        TaskSetupperBase.__init__(self,taskBufferIF,ddmIF)


    # main to setup task
    def doSetup(self,taskSpec,datasetToRegister,pandaJobs):
        return self.SC_SUCCEEDED
            




from pandacommon.pandamsgbkr.msg_processor import SimpleMsgProcPluginBase

from pandajedi.jedicore.JediTaskBuffer import JediTaskBuffer

# Base simple message processing plugin
class BaseMsgProcPlugin(SimpleMsgProcPluginBase):

    def initialize(self):
        """
        initialize plugin instance, run once before loop in thread
        """
        # set up JEDI TaskBuffer interface
        self.tbIF= JediTaskBuffer(None)

    def process(self, msg_obj):
        """
        process the message
        Get msg_obj from the incoming MQ (if any; otherwise msg_obj is None)
        Returned value will be sent to the outgoing MQ (if any)
        """
        pass



# DB API for JEDI

import datetime

from . import JediDBProxyPool
from .Interaction import CommandReceiveInterface
from pandajedi.jediconfig import jedi_config

from pandaserver.taskbuffer import TaskBuffer
from pandaserver.brokerage.SiteMapper import SiteMapper


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])

# use customized proxy pool
TaskBuffer.DBProxyPool = JediDBProxyPool.DBProxyPool


class JediTaskBuffer(TaskBuffer.TaskBuffer, CommandReceiveInterface):

    # constructor
    def __init__(self,conn):
        CommandReceiveInterface.__init__(self,conn)
        TaskBuffer.TaskBuffer.__init__(self)
        TaskBuffer.TaskBuffer.init(self, jedi_config.db.dbhost,
                                   jedi_config.db.dbpasswd,
                                   nDBConnection=1)
        # site mapper
        self.siteMapper = SiteMapper(self)
        # update time for site mapper
        self.dateTimeForSM = datetime.datetime.utcnow()
        logger.debug('__init__')

    # query an SQL
    def querySQL(self, sql, varMap, arraySize=1000):
        with self.proxyPool.get() as proxy:
            return proxy.querySQLS(sql, varMap, arraySize)[1]

    # get SiteMapper
    def getSiteMapper(self):
        timeNow = datetime.datetime.utcnow()
        if datetime.datetime.utcnow()-self.dateTimeForSM > datetime.timedelta(minutes=10):
            self.siteMapper = SiteMapper(self)
            self.dateTimeForSM = timeNow
        return self.siteMapper

    # get work queue map
    def getWorkQueueMap(self):
        with self.proxyPool.get() as proxy:
            return proxy.getWorkQueueMap()

    # get the list of datasets to feed contents to DB
    def getDatasetsToFeedContents_JEDI(self,vo=None,prodSourceLabel=None):
        with self.proxyPool.get() as proxy:
            return proxy.getDatasetsToFeedContents_JEDI(vo,prodSourceLabel)

    # feed files to the JEDI contents table
    def insertFilesForDataset_JEDI(self,datasetSpec,fileMap,datasetState,stateUpdateTime,
                                   nEventsPerFile,nEventsPerJob,maxAttempt,firstEventNumber,
                                   nMaxFiles,nMaxEvents,useScout,fileList,useFilesWithNewAttemptNr,
                                   nFilesPerJob,nEventsPerRange,nChunksForScout,includePatt,
                                   excludePatt,xmlConfig,noWaitParent,parent_tid,pid,maxFailure,
                                   useRealNumEvents,respectLB,tgtNumEventsPerJob,skipFilesUsedBy,
                                   ramCount,taskSpec,skipShortInput, inputPreStaging):
        with self.proxyPool.get() as proxy:
            return proxy.insertFilesForDataset_JEDI(datasetSpec,fileMap,datasetState,stateUpdateTime,
                                                  nEventsPerFile,nEventsPerJob,maxAttempt,
                                                  firstEventNumber,nMaxFiles,nMaxEvents,
                                                  useScout,fileList,useFilesWithNewAttemptNr,
                                                  nFilesPerJob,nEventsPerRange,nChunksForScout,
                                                  includePatt,excludePatt,xmlConfig,
                                                  noWaitParent,parent_tid,pid,maxFailure,
                                                  useRealNumEvents,respectLB,
                                                  tgtNumEventsPerJob,skipFilesUsedBy,
                                                  ramCount,taskSpec,skipShortInput, inputPreStaging)

    # get files from the JEDI contents table with jediTaskID and/or datasetID
    def getFilesInDatasetWithID_JEDI(self,jediTaskID=None,datasetID=None,nFiles=None,status=None):
        with self.proxyPool.get() as proxy:
            return proxy.getFilesInDatasetWithID_JEDI(jediTaskID,datasetID,nFiles,status)

    # insert dataset to the JEDI datasets table
    def insertDataset_JEDI(self,datasetSpec):
        with self.proxyPool.get() as proxy:
            return proxy.insertDataset_JEDI(datasetSpec)

    # update JEDI dataset
    def updateDataset_JEDI(self,datasetSpec,criteria,lockTask=False):
        with self.proxyPool.get() as proxy:
            return proxy.updateDataset_JEDI(datasetSpec,criteria,lockTask)

    # update JEDI dataset attributes
    def updateDatasetAttributes_JEDI(self,jediTaskID,datasetID,attributes):
        with self.proxyPool.get() as proxy:
            return proxy.updateDatasetAttributes_JEDI(jediTaskID,datasetID,attributes)

    # get JEDI dataset attributes
    def getDatasetAttributes_JEDI(self,jediTaskID,datasetID,attributes):
        with self.proxyPool.get() as proxy:
            return proxy.getDatasetAttributes_JEDI(jediTaskID,datasetID,attributes)

    # get JEDI dataset attributes with map
    def getDatasetAttributesWithMap_JEDI(self,jediTaskID,criteria,attributes):
        with self.proxyPool.get() as proxy:
            return proxy.getDatasetAttributesWithMap_JEDI(jediTaskID,criteria,attributes)

    # get JEDI dataset with jediTaskID and datasetID
    def getDatasetWithID_JEDI(self,jediTaskID,datasetID):
        with self.proxyPool.get() as proxy:
            return proxy.getDatasetWithID_JEDI(jediTaskID,datasetID)

    # get JEDI datasets with jediTaskID
    def getDatasetsWithJediTaskID_JEDI(self,jediTaskID,datasetTypes=None,getFiles=False):
        with self.proxyPool.get() as proxy:
            retStat, datasetSpecList = proxy.getDatasetsWithJediTaskID_JEDI(jediTaskID, datasetTypes=datasetTypes)
            if retStat is True and getFiles is True:
                for datasetSpec in datasetSpecList:
                    # read files
                    retStat, fileSpecList = proxy.getFilesInDatasetWithID_JEDI(jediTaskID, datasetSpec.datasetID, None, None)
                    if retStat is False:
                        break
                    for fileSpec in fileSpecList:
                        datasetSpec.addFile(fileSpec)
            # return
            return retStat, datasetSpecList

    # insert task to the JEDI tasks table
    def insertTask_JEDI(self,taskSpec):
        with self.proxyPool.get() as proxy:
            return proxy.insertTask_JEDI(taskSpec)

    # update JEDI task
    def updateTask_JEDI(self,taskSpec,criteria,oldStatus=None,updateDEFT=False,insertUnknown=None,
                        setFrozenTime=True,setOldModTime=False):
        with self.proxyPool.get() as proxy:
            return proxy.updateTask_JEDI(taskSpec,criteria,oldStatus,updateDEFT,insertUnknown,
                                       setFrozenTime,setOldModTime)

    # update JEDI task lock
    def updateTaskLock_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.updateTaskLock_JEDI(jediTaskID)

    # update JEDI task status by ContentsFeeder
    def updateTaskStatusByContFeeder_JEDI(self,jediTaskID,taskSpec=None,getTaskStatus=False,pid=None,
                                          setFrozenTime=True,useWorldCloud=False):
        with self.proxyPool.get() as proxy:
            return proxy.updateTaskStatusByContFeeder_JEDI(jediTaskID,taskSpec,getTaskStatus,
                                                         pid,setFrozenTime,useWorldCloud)

    # get JEDI task with jediTaskID
    def getTaskWithID_JEDI(self, jediTaskID, fullFlag=False, lockTask=False, pid=None, lockInterval=None,
                           clearError=False):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskWithID_JEDI(jediTaskID, fullFlag, lockTask, pid, lockInterval, clearError)

    # get JEDI task and tasks with ID and lock it
    def getTaskDatasetsWithID_JEDI(self,jediTaskID,pid,lockTask=True):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskDatasetsWithID_JEDI(jediTaskID,pid,lockTask)

    # get JEDI tasks with selection criteria
    def getTaskIDsWithCriteria_JEDI(self,criteria,nTasks=50):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskIDsWithCriteria_JEDI(criteria,nTasks)

    # get JEDI tasks to be finished
    def getTasksToBeFinished_JEDI(self,vo,prodSourceLabel,pid,nTasks=50):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToBeFinished_JEDI(vo,prodSourceLabel,pid,nTasks)

    # get tasks to be processed
    def getTasksToBeProcessed_JEDI(self,pid,vo,workQueue,prodSourceLabel,cloudName,
                                   nTasks=50,nFiles=100,simTasks=None,minPriority=None,
                                   maxNumJobs=None,typicalNumFilesMap=None,
                                   fullSimulation=False,simDatasets=None,
                                   mergeUnThrottled=None,readMinFiles=False,
                                   numNewTaskWithJumbo=0, resource_name=None,
                                   ignore_lock=False, target_tasks=None):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToBeProcessed_JEDI(pid,vo,workQueue,prodSourceLabel,cloudName,nTasks,nFiles,
                                                  simTasks=simTasks,
                                                  minPriority=minPriority,
                                                  maxNumJobs=maxNumJobs,
                                                  typicalNumFilesMap=typicalNumFilesMap,
                                                  fullSimulation=fullSimulation,
                                                  simDatasets=simDatasets,
                                                  mergeUnThrottled=mergeUnThrottled,
                                                  readMinFiles=readMinFiles,
                                                  numNewTaskWithJumbo=numNewTaskWithJumbo,
                                                  resource_name=resource_name,
                                                  ignore_lock=ignore_lock,
                                                  target_tasks=target_tasks)

    # get tasks to be processed
    def checkWaitingTaskPrio_JEDI(self, vo, workQueue, prodSourceLabel, cloudName, resource_name=None):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToBeProcessed_JEDI(None, vo, workQueue, prodSourceLabel,
                                                  cloudName, isPeeking=True, resource_name=resource_name)

    # get job statistics with work queue
    def getJobStatisticsWithWorkQueue_JEDI(self,vo,prodSourceLabel,minPriority=None):
        with self.proxyPool.get() as proxy:
            return proxy.getJobStatisticsWithWorkQueue_JEDI(vo,prodSourceLabel,minPriority)

    # get job statistics by global share
    def getJobStatisticsByGlobalShare(self, vo, exclude_rwq=False):
        with self.proxyPool.get() as proxy:
            return proxy.getJobStatisticsByGlobalShare(vo, exclude_rwq)

    # get job statistics by global share
    def getJobStatisticsByResourceType(self, workqueue):
        with self.proxyPool.get() as proxy:
            return proxy.getJobStatisticsByResourceType(workqueue)

    # get job statistics by site and resource type
    def getJobStatisticsByResourceTypeSite(self, workqueue):
        with self.proxyPool.get() as proxy:
            return proxy.getJobStatisticsByResourceTypeSite(workqueue)

    # generate output files for task
    def getOutputFiles_JEDI(self,jediTaskID,provenanceID,simul,instantiateTmpl=False,instantiatedSite=None,
                            isUnMerging=False,isPrePro=False,xmlConfigJob=None,siteDsMap=None,middleName='',
                            registerDatasets=False,parallelOutMap=None,fileIDPool=[]):
        with self.proxyPool.get() as proxy:
            return proxy.getOutputFiles_JEDI(jediTaskID,provenanceID,simul,instantiateTmpl,instantiatedSite,
                                           isUnMerging,isPrePro,xmlConfigJob,siteDsMap,middleName,
                                           registerDatasets,parallelOutMap,fileIDPool)

    # insert output file templates
    def insertOutputTemplate_JEDI(self,templates):
        with self.proxyPool.get() as proxy:
            return proxy.insertOutputTemplate_JEDI(templates)

    # insert JobParamsTemplate
    def insertJobParamsTemplate_JEDI(self,jediTaskID,templ):
        with self.proxyPool.get() as proxy:
            return proxy.insertJobParamsTemplate_JEDI(jediTaskID,templ)

    # insert TaskParams
    def insertTaskParams_JEDI(self,vo,prodSourceLabel,userName,taskName,taskParams,parent_tid=None):
        with self.proxyPool.get() as proxy:
            return proxy.insertTaskParams_JEDI(vo,prodSourceLabel,userName,taskName,taskParams,parent_tid)

    # reset unused files
    def resetUnusedFiles_JEDI(self,jediTaskID,inputChunk):
        with self.proxyPool.get() as proxy:
            return proxy.resetUnusedFiles_JEDI(jediTaskID,inputChunk)

    # insert TaskParams
    def insertUpdateTaskParams_JEDI(self,jediTaskID,vo,prodSourceLabel,updateTaskParams,insertTaskParamsList):
        with self.proxyPool.get() as proxy:
            return proxy.insertUpdateTaskParams_JEDI(jediTaskID,vo,prodSourceLabel,updateTaskParams,insertTaskParamsList)

    # set missing files
    def setMissingFiles_JEDI(self,jediTaskID,datasetID,fileIDs):
        with self.proxyPool.get() as proxy:
            return proxy.setMissingFiles_JEDI(jediTaskID,datasetID,fileIDs)

    # rescue picked files
    def rescuePickedFiles_JEDI(self,vo,prodSourceLabel,waitTime):
        with self.proxyPool.get() as proxy:
            return proxy.rescuePickedFiles_JEDI(vo,prodSourceLabel,waitTime)

    # rescue unlocked tasks with picked files
    def rescueUnLockedTasksWithPicked_JEDI(self,vo,prodSourceLabel,waitTime,pid):
        with self.proxyPool.get() as proxy:
            return proxy.rescueUnLockedTasksWithPicked_JEDI(vo,prodSourceLabel,waitTime,pid)

    # unlock tasks
    def unlockTasks_JEDI(self,vo,prodSourceLabel,waitTime,hostName=None,pgid=None):
        with self.proxyPool.get() as proxy:
            return proxy.unlockTasks_JEDI(vo,prodSourceLabel,waitTime,hostName,pgid)

    # get the size of input files which will be copied to the site
    def getMovingInputSize_JEDI(self,siteName):
        with self.proxyPool.get() as proxy:
            return proxy.getMovingInputSize_JEDI(siteName)

    # get typical number of input files for each workQueue+processingType
    def getTypicalNumInput_JEDI(self,vo,prodSourceLabel,workQueue):
        with self.proxyPool.get() as proxy:
            return proxy.getTypicalNumInput_JEDI(vo,prodSourceLabel,workQueue)

    # get highest prio jobs with workQueueID
    def getHighestPrioJobStat_JEDI(self, prodSourceLabel, cloudName, workQueue, resource_name=None):
        with self.proxyPool.get() as proxy:
            return proxy.getHighestPrioJobStat_JEDI(prodSourceLabel, cloudName, workQueue, resource_name)

    # get the list of tasks to refine
    def getTasksToRefine_JEDI(self,vo=None,prodSourceLabel=None):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToRefine_JEDI(vo,prodSourceLabel)

    # get task parameters with jediTaskID
    def getTaskParamsWithID_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskParamsWithID_JEDI(jediTaskID)

    # register task/dataset/templ/param in a single transaction
    def registerTaskInOneShot_JEDI(self,jediTaskID,taskSpec,inMasterDatasetSpec,
                                   inSecDatasetSpecList,outDatasetSpecList,
                                   outputTemplateMap,jobParamsTemplate,taskParams,
                                   unmergeMasterDatasetSpec,unmergeDatasetSpecMap,
                                   uniqueTaskName,oldTaskStatus):
        with self.proxyPool.get() as proxy:
            return proxy.registerTaskInOneShot_JEDI(jediTaskID,taskSpec,inMasterDatasetSpec,
                                                  inSecDatasetSpecList,outDatasetSpecList,
                                                  outputTemplateMap,jobParamsTemplate,taskParams,
                                                  unmergeMasterDatasetSpec,unmergeDatasetSpecMap,
                                                  uniqueTaskName,oldTaskStatus)

    # set tasks to be assigned
    def setScoutJobDataToTasks_JEDI(self,vo,prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.setScoutJobDataToTasks_JEDI(vo, prodSourceLabel, self.siteMapper)

    # prepare tasks to be finished
    def prepareTasksToBeFinished_JEDI(self,vo,prodSourceLabel,nTasks=50,simTasks=None,pid='lock',noBroken=False):
        with self.proxyPool.get() as proxy:
            return proxy.prepareTasksToBeFinished_JEDI(vo,prodSourceLabel,nTasks,simTasks,pid,noBroken,
                                                       self.siteMapper)

    # get tasks to be assigned
    def getTasksToAssign_JEDI(self, vo, prodSourceLabel, workQueue, resource_name):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToAssign_JEDI(vo, prodSourceLabel, workQueue, resource_name)

    # get tasks to check task assignment
    def getTasksToCheckAssignment_JEDI(self, vo, prodSourceLabel, workQueue, resource_name):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToCheckAssignment_JEDI(vo, prodSourceLabel, workQueue, resource_name)

    # calculate RW with a priority
    def calculateRWwithPrio_JEDI(self,vo,prodSourceLabel,workQueue,priority):
        with self.proxyPool.get() as proxy:
            return proxy.calculateRWwithPrio_JEDI(vo,prodSourceLabel,workQueue,priority)

    # calculate RW for tasks
    def calculateTaskRW_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.calculateTaskRW_JEDI(jediTaskID)

    # calculate WORLD RW with a priority
    def calculateWorldRWwithPrio_JEDI(self,vo,prodSourceLabel,workQueue,priority):
        with self.proxyPool.get() as proxy:
            return proxy.calculateWorldRWwithPrio_JEDI(vo,prodSourceLabel,workQueue,priority)

    # calculate WORLD RW for tasks
    def calculateTaskWorldRW_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.calculateTaskWorldRW_JEDI(jediTaskID)

    # set cloud to tasks
    def setCloudToTasks_JEDI(self,taskCloudMap):
        with self.proxyPool.get() as proxy:
            return proxy.setCloudToTasks_JEDI(taskCloudMap)

    # get the list of tasks to exec command
    def getTasksToExecCommand_JEDI(self,vo,prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToExecCommand_JEDI(vo,prodSourceLabel)

    # get the list of PandaIDs for a task
    def getPandaIDsWithTask_JEDI(self,jediTaskID,onlyActive):
        with self.proxyPool.get() as proxy:
            return proxy.getPandaIDsWithTask_JEDI(jediTaskID,onlyActive)

    # get the list of queued PandaIDs for a task
    def getQueuedPandaIDsWithTask_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getQueuedPandaIDsWithTask_JEDI(jediTaskID)

    # get jediTaskID/datasetID/FileID with dataset and file names
    def getIDsWithFileDataset_JEDI(self,datasetName,fileName,fileType):
        with self.proxyPool.get() as proxy:
            return proxy.getIDsWithFileDataset_JEDI(datasetName,fileName,fileType)

    # get PandaID for a file
    def getPandaIDWithFileID_JEDI(self,jediTaskID,datasetID,fileID):
        with self.proxyPool.get() as proxy:
            return proxy.getPandaIDWithFileID_JEDI(jediTaskID,datasetID,fileID)

    # get JEDI files for a job
    def getFilesWithPandaID_JEDI(self,pandaID):
        with self.proxyPool.get() as proxy:
            return proxy.getFilesWithPandaID_JEDI(pandaID)

    # update task parameters
    def updateTaskParams_JEDI(self,jediTaskID,taskParams):
        with self.proxyPool.get() as proxy:
            return proxy.updateTaskParams_JEDI(jediTaskID,taskParams)

    # reactivate pending tasks
    def reactivatePendingTasks_JEDI(self,vo,prodSourceLabel,timeLimit,timeoutLimit=None,minPriority=None):
        with self.proxyPool.get() as proxy:
            return proxy.reactivatePendingTasks_JEDI(vo,prodSourceLabel,timeLimit,timeoutLimit,minPriority)

    # restart contents update
    def restartTasksForContentsUpdate_JEDI(self,vo,prodSourceLabel,timeLimit=30):
        with self.proxyPool.get() as proxy:
            return proxy.restartTasksForContentsUpdate_JEDI(vo,prodSourceLabel,timeLimit=timeLimit)

    # kick exhausted tasks
    def kickExhaustedTasks_JEDI(self,vo,prodSourceLabel,timeLimit):
        with self.proxyPool.get() as proxy:
            return proxy.kickExhaustedTasks_JEDI(vo,prodSourceLabel,timeLimit)

    # get file spec of lib.tgz
    def getBuildFileSpec_JEDI(self,jediTaskID,siteName,associatedSites):
        with self.proxyPool.get() as proxy:
            return proxy.getBuildFileSpec_JEDI(jediTaskID,siteName,associatedSites)

    # get file spec of old lib.tgz
    def getOldBuildFileSpec_JEDI(self,jediTaskID,datasetID,fileID):
        with self.proxyPool.get() as proxy:
            return proxy.getOldBuildFileSpec_JEDI(jediTaskID,datasetID,fileID)

    # insert lib dataset and files
    def insertBuildFileSpec_JEDI(self,jobSpec,reusedDatasetID,simul):
        with self.proxyPool.get() as proxy:
            return proxy.insertBuildFileSpec_JEDI(jobSpec,reusedDatasetID,simul)

    # get sites used by a task
    def getSitesUsedByTask_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getSitesUsedByTask_JEDI(jediTaskID)

    # get random seed
    def getRandomSeed_JEDI(self,jediTaskID,simul):
        with self.proxyPool.get() as proxy:
            return proxy.getRandomSeed_JEDI(jediTaskID,simul)

    # get preprocess metadata
    def getPreprocessMetadata_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getPreprocessMetadata_JEDI(jediTaskID)

    # get log dataset for preprocessing
    def getPreproLog_JEDI(self,jediTaskID,simul):
        with self.proxyPool.get() as proxy:
            return proxy.getPreproLog_JEDI(jediTaskID,simul)

    # get sites with best connections to source
    def getBestNNetworkSites_JEDI(self,source,protocol,nSites,threshold,cutoff,maxWeight):
        with self.proxyPool.get() as proxy:
            return proxy.getBestNNetworkSites_JEDI(source,protocol,nSites,threshold,
                                                 cutoff,maxWeight)

    # get jobsetID
    def getUserJobsetID_JEDI(self,userName):
        with self.proxyPool.get() as proxy:
            tmpJobID, tmpDummy, tmpStat = proxy.getUserParameter(userName, 1, None)
            # return
            return tmpStat, tmpJobID

    # retry or incrementally execute a task
    def retryTask_JEDI(self,jediTaskID,commStr,maxAttempt=5,retryChildTasks=True,discardEvents=False,
                       release_unstaged=False):
        with self.proxyPool.get() as proxy:
            return proxy.retryTask_JEDI(jediTaskID,commStr,maxAttempt,retryChildTasks=retryChildTasks,
                                      discardEvents=discardEvents, release_unstaged=release_unstaged)

    # append input datasets for incremental execution
    def appendDatasets_JEDI(self,jediTaskID,inMasterDatasetSpecList,inSecDatasetSpecList):
        with self.proxyPool.get() as proxy:
            return proxy.appendDatasets_JEDI(jediTaskID,inMasterDatasetSpecList,inSecDatasetSpecList)

    # record retry history
    def recordRetryHistory_JEDI(self,jediTaskID,oldNewPandaIDs,relationType):
        with self.proxyPool.get() as proxy:
            return proxy.recordRetryHistory_JEDI(jediTaskID,oldNewPandaIDs,relationType)

    # get JEDI tasks with a selection criteria
    def getTasksWithCriteria_JEDI(self,vo,prodSourceLabel,taskStatusList,taskCriteria={},datasetCriteria={},
                                  taskParamList=[],datasetParamList=[],taskLockColumn=None,taskLockInterval=60):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksWithCriteria_JEDI(vo,prodSourceLabel,taskStatusList,taskCriteria,datasetCriteria,
                                                 taskParamList,datasetParamList,taskLockColumn,taskLockInterval)

    # check parent task status
    def checkParentTask_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.checkParentTask_JEDI(jediTaskID)

    # get task status
    def getTaskStatus_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskStatus_JEDI(jediTaskID)

    # get lib.tgz for waiting jobs
    def getLibForWaitingRunJob_JEDI(self,vo,prodSourceLabel,checkInterval):
        with self.proxyPool.get() as proxy:
            return proxy.getLibForWaitingRunJob_JEDI(vo,prodSourceLabel,checkInterval)

    # get tasks to get reassigned
    def getTasksToReassign_JEDI(self,vo=None,prodSourceLabel=None):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToReassign_JEDI(vo,prodSourceLabel)

    # kill child tasks
    def killChildTasks_JEDI(self,jediTaskID,taskStatus):
        with self.proxyPool.get() as proxy:
            return proxy.killChildTasks_JEDI(jediTaskID,taskStatus)

    # kick child tasks
    def kickChildTasks_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.kickChildTasks_JEDI(jediTaskID)

    # lock task
    def lockTask_JEDI(self,jediTaskID,pid):
        with self.proxyPool.get() as proxy:
            return proxy.lockTask_JEDI(jediTaskID,pid)

    # get successful files
    def getSuccessfulFiles_JEDI(self,jediTaskID,datasetID):
        with self.proxyPool.get() as proxy:
            return proxy.getSuccessfulFiles_JEDI(jediTaskID,datasetID)

    # unlock a single task
    def unlockSingleTask_JEDI(self,jediTaskID,pid):
        with self.proxyPool.get() as proxy:
            return proxy.unlockSingleTask_JEDI(jediTaskID,pid)

    # throttle JEDI tasks
    def throttleTasks_JEDI(self,vo,prodSourceLabel,waitTime):
        with self.proxyPool.get() as proxy:
            return proxy.throttleTasks_JEDI(vo,prodSourceLabel,waitTime)

    # throttle a JEDI task
    def throttleTask_JEDI(self,jediTaskID,waitTime,errorDialog):
        with self.proxyPool.get() as proxy:
            return proxy.throttleTask_JEDI(jediTaskID,waitTime,errorDialog)

    # release throttled tasks
    def releaseThrottledTasks_JEDI(self,vo,prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.releaseThrottledTasks_JEDI(vo,prodSourceLabel)

    # release throttled task
    def releaseThrottledTask_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.releaseThrottledTask_JEDI(jediTaskID)

    # get throttled users
    def getThrottledUsersTasks_JEDI(self,vo,prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.getThrottledUsersTasks_JEDI(vo,prodSourceLabel)

    # lock process
    def lockProcess_JEDI(self, vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid, forceOption=False, timeLimit=5):
        with self.proxyPool.get() as proxy:
            return proxy.lockProcess_JEDI(vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid, forceOption, timeLimit)

    # unlock process
    def unlockProcess_JEDI(self, vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid):
        with self.proxyPool.get() as proxy:
            return proxy.unlockProcess_JEDI(vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid)

    # unlock process with PID
    def unlockProcessWithPID_JEDI(self, vo, prodSourceLabel, workqueue_id, resource_name, pid, useBase):
        with self.proxyPool.get() as proxy:
            return proxy.unlockProcessWithPID_JEDI(vo, prodSourceLabel, workqueue_id, resource_name, pid, useBase)

    # check process lock
    def checkProcessLock_JEDI(self, vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid, checkBase):
        with self.proxyPool.get() as proxy:
            return proxy.checkProcessLock_JEDI(vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid, checkBase)

    # get JEDI tasks to be assessed
    def getAchievedTasks_JEDI(self,vo,prodSourceLabel,timeLimit=60,nTasks=50):
        with self.proxyPool.get() as proxy:
            return proxy.getAchievedTasks_JEDI(vo,prodSourceLabel,timeLimit,nTasks)

    # get inactive sites
    def getInactiveSites_JEDI(self,flag,timeLimit):
        with self.proxyPool.get() as proxy:
            return proxy.getInactiveSites_JEDI(flag,timeLimit)

    # get total walltime
    def getTotalWallTime_JEDI(self, vo, prodSourceLabel, workQueue, resource_name):
        with self.proxyPool.get() as proxy:
            return proxy.getTotalWallTime_JEDI(vo, prodSourceLabel, workQueue, resource_name)

    # check duplication with internal merge
    def checkDuplication_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.checkDuplication_JEDI(jediTaskID)

    # get network metrics for brokerage
    def getNetworkMetrics(self, dst, keyList):
        with self.proxyPool.get() as proxy:
            return proxy.getNetworkMetrics(dst, keyList)

    # get nuclei that have built up a long backlog
    def getBackloggedNuclei(self):
        with self.proxyPool.get() as proxy:
            return proxy.getBackloggedNuclei()

    # get network metrics for brokerage
    def getPandaSiteToOutputStorageSiteMapping(self):
        with self.proxyPool.get() as proxy:
            return proxy.getPandaSiteToOutputStorageSiteMapping()

    # get failure counts for a task
    def getFailureCountsForTask_JEDI(self,jediTaskID,timeWindow):
        with self.proxyPool.get() as proxy:
            return proxy.getFailureCountsForTask_JEDI(jediTaskID,timeWindow)

        # count the number of queued jobs per user or working group
    def countJobsPerTarget_JEDI(self, target, is_user):
        with self.proxyPool.get() as proxy:
            return proxy.countJobsPerTarget_JEDI(target, is_user)

    # get old merge job PandaIDs
    def getOldMergeJobPandaIDs_JEDI(self,jediTaskID,pandaID):
        with self.proxyPool.get() as proxy:
            return proxy.getOldMergeJobPandaIDs_JEDI(jediTaskID,pandaID)

    # get active jumbo jobs for a task
    def getActiveJumboJobs_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getActiveJumboJobs_JEDI(jediTaskID)

    # get jobParms of the first job
    def getJobParamsOfFirstJob_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getJobParamsOfFirstJob_JEDI(jediTaskID)

    # bulk fetch fileIDs
    def bulkFetchFileIDs_JEDI(self,jediTaskID,nIDs):
        with self.proxyPool.get() as proxy:
            return proxy.bulkFetchFileIDs_JEDI(jediTaskID,nIDs)

    # set del flag to events
    def setDelFlagToEvents_JEDI(self,jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.setDelFlagToEvents_JEDI(jediTaskID)

    # set del flag to events
    def removeFilesIndexInconsistent_JEDI(self,jediTaskID,datasetIDs):
        with self.proxyPool.get() as proxy:
            return proxy.removeFilesIndexInconsistent_JEDI(jediTaskID,datasetIDs)

    # throttle jobs in pauses tasks
    def throttleJobsInPausedTasks_JEDI(self,vo,prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.throttleJobsInPausedTasks_JEDI(vo,prodSourceLabel)

    # set useJumbo flag
    def setUseJumboFlag_JEDI(self,jediTaskID,statusStr):
        with self.proxyPool.get() as proxy:
            return proxy.setUseJumboFlag_JEDI(jediTaskID,statusStr)

    # get number of tasks with running jumbo jobs
    def getNumTasksWithRunningJumbo_JEDI(self,vo,prodSourceLabel,cloudName,workqueue):
        with self.proxyPool.get() as proxy:
            return proxy.getNumTasksWithRunningJumbo_JEDI(vo,prodSourceLabel,cloudName,workqueue)

    # get number of unprocessed events
    def getNumUnprocessedEvents_JEDI(self, vo, prodSourceLabel, criteria, neg_criteria):
        with self.proxyPool.get() as proxy:
            return proxy.getNumUnprocessedEvents_JEDI(vo, prodSourceLabel, criteria, neg_criteria)

    # get number of jobs for a task
    def getNumJobsForTask_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getNumJobsForTask_JEDI(jediTaskID)

    # get number map for standby jobs
    def getNumMapForStandbyJobs_JEDI(self, workqueue):
        with self.proxyPool.get() as proxy:
            return proxy.getNumMapForStandbyJobs_JEDI(workqueue)

    # update datasets to finish a task
    def updateDatasetsToFinishTask_JEDI(self, jediTaskID, lockedBy):
        with self.proxyPool.get() as proxy:
            return proxy.updateDatasetsToFinishTask_JEDI(jediTaskID, lockedBy)

    # get tasks with jumbo jobs
    def getTaskWithJumbo_JEDI(self, vo, prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskWithJumbo_JEDI(vo, prodSourceLabel)

    # get tasks to enabme jumbo jobs
    def getTaskToEnableJumbo_JEDI(self, vo, prodSourceLabel, maxPrio, nEventsToEnable):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskToEnableJumbo_JEDI(vo, prodSourceLabel, maxPrio, nEventsToEnable)

    # get release and cache at jumbo job enabled sites
    def getRelCacheForJumbo_JEDI(self):
        with self.proxyPool.get() as proxy:
            return proxy.getRelCacheForJumbo_JEDI()

    # kick pending tasks with jumbo jobs
    def kickPendingTasksWithJumbo_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.kickPendingTasksWithJumbo_JEDI(jediTaskID)

    # reset input to re-generate co-jumbo jobs
    def resetInputToReGenCoJumbo_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.resetInputToReGenCoJumbo_JEDI(jediTaskID)

    # get averaged disk IO
    def getAvgDiskIO_JEDI(self):
        with self.proxyPool.get() as proxy:
            return proxy.getAvgDiskIO_JEDI()

    # update input files stage-in done according to message from idds
    def updateInputFilesStagedAboutIdds_JEDI(self, jeditaskid, scope, filenames):
        with self.proxyPool.get() as proxy:
            return proxy.updateInputFilesStagedAboutIdds_JEDI(jeditaskid, scope, filenames)

    # update input datasets stage-in done according to message from idds
    def updateInputDatasetsStagedAboutIdds_JEDI(self, jeditaskid, scope, dsnames):
        with self.proxyPool.get() as proxy:
            return proxy.updateInputDatasetsStagedAboutIdds_JEDI(jeditaskid, scope, dsnames)

   # get number of staging files
    def getNumStagingFiles_JEDI(self, jeditaskid):
        with self.proxyPool.get() as proxy:
            return proxy.getNumStagingFiles_JEDI(jeditaskid)

    # get usage breakdown by users and sites
    def getUsageBreakdown_JEDI(self, prod_source_label='user'):
        with self.proxyPool.get() as proxy:
            return proxy.getUsageBreakdown_JEDI(prod_source_label)

    # get jobs stat of each user
    def getUsersJobsStats_JEDI(self, prod_source_label='user'):
        with self.proxyPool.get() as proxy:
            return proxy.getUsersJobsStats_JEDI(prod_source_label)

    # insert HPO pseudo event according to message from idds
    def insertHpoEventAboutIdds_JEDI(self, jedi_task_id, event_id_list):
        with self.proxyPool.get() as proxy:
            return proxy.insertHpoEventAboutIdds_JEDI(jedi_task_id, event_id_list)

    # get event statistics
    def get_event_statistics(self, jedi_task_id):
        with self.proxyPool.get() as proxy:
            return proxy.get_event_statistics(jedi_task_id)

    # get site to-running rate statistics by global share
    def getSiteToRunRateStats(self, vo, exclude_rwq, starttime_min, starttime_max):
        with self.proxyPool.get() as proxy:
            return proxy.getSiteToRunRateStats(vo, exclude_rwq, starttime_min, starttime_max)

    # update cache
    def updateCache_JEDI(self, main_key, sub_key, data):
        with self.proxyPool.get() as proxy:
            return proxy.updateCache_JEDI(main_key, sub_key, data)

    # get cache
    def getCache_JEDI(self, main_key, sub_key):
        with self.proxyPool.get() as proxy:
            return proxy.getCache_JEDI(main_key, sub_key)

    # get cache
    def extendSandboxLifetime_JEDI(self, jedi_taskid, file_name):
        with self.proxyPool.get() as proxy:
            return proxy.extendSandboxLifetime_JEDI(jedi_taskid, file_name)

    # turn a task into pending status for some reason
    def makeTaskPending_JEDI(self, jedi_taskid, reason='unknown'):
        with self.proxyPool.get() as proxy:
            return proxy.makeTaskPending_JEDI(jedi_taskid, reason)

    # query tasks and turn them into pending status for some reason, sql_query should query jeditaskid
    def queryTasksToBePending_JEDI(self, sql_query, params_map, reason):
        with self.proxyPool.get() as proxy:
            return proxy.queryTasksToBePending_JEDI(sql_query, params_map, reason)

    # get IDs of all datasets of input and lib, to update data locality records
    def get_tasks_inputdatasets_JEDI(self, vo):
        with self.proxyPool.get() as proxy:
            return proxy.get_tasks_inputdatasets_JEDI(vo)

    # update dataset locality
    def updateDatasetLocality_JEDI(self, jedi_taskid, datasetid, rse):
        with self.proxyPool.get() as proxy:
            return proxy.updateDatasetLocality_JEDI(jedi_taskid, datasetid, rse)

    # delete outdated dataset locality records
    def deleteOutdatedDatasetLocality_JEDI(self, before_timestamp):
        with self.proxyPool.get() as proxy:
            return proxy.deleteOutdatedDatasetLocality_JEDI(before_timestamp)

    # query tasks and preassign them to dedicate workqueue, sql_query should query jeditaskid
    def queryTasksToPreassign_JEDI(self, sql_query, params_map, site, blacklist=set(), limit=1):
        with self.proxyPool.get() as proxy:
            return proxy.queryTasksToPreassign_JEDI(sql_query, params_map, site, blacklist, limit)

    # close and reassign N jobs of a preassigned task
    def reassignJobsInPreassignedTask_JEDI(self, jedi_taskid, site, n_jobs_to_close):
        with self.proxyPool.get() as proxy:
            return proxy.reassignJobsInPreassignedTask_JEDI(jedi_taskid, site, n_jobs_to_close)

    # undo preassigned tasks
    def undoPreassignedTasks_JEDI(self, jedi_taskids, task_orig_attr_map, params_map, force=False):
        with self.proxyPool.get() as proxy:
            return proxy.undoPreassignedTasks_JEDI(jedi_taskids, task_orig_attr_map, params_map, force)

    # set missing files according to iDDS messages
    def setMissingFilesAboutIdds_JEDI(self, jeditaskid, filenames):
        with self.proxyPool.get() as proxy:
            return proxy.setMissingFilesAboutIdds_JEDI(jeditaskid, filenames)



import sys
try:
    metaID = sys.argv[1]
except Exception:
    metaID = None
import json
import uuid

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

reqIdx = 100000000
dsName = 'mc12_8TeV.129933.McAtNloJimmy_AUET2CT10_WW_e_e_0_0_0_0_0.merge.AOD.e1563_s1499_s1504_r3658_r3549_tid00999168_00'
logDatasetName = 'panda.jeditest.log.{0}'.format(uuid.uuid4())

taskParamMap = {}

taskParamMap['nFilesPerJob'] = 1
taskParamMap['nFiles'] = 2
taskParamMap['nEventsPerInputFile']  = 10000
taskParamMap['nEventsPerJob'] = 10000
#taskParamMap['nEvents'] = 25000
#taskParamMap['skipScout'] = True
taskParamMap['taskName'] = str(uuid.uuid4())
taskParamMap['userName'] = 'pandasrv1'
taskParamMap['vo'] = 'atlas'
taskParamMap['taskPriority'] = 900
#taskParamMap['reqID'] = reqIdx
taskParamMap['architecture'] = 'i686-slc5-gcc43-opt'
taskParamMap['transUses'] = 'Atlas-17.2.2'
taskParamMap['transHome'] = 'AtlasProduction-17.2.2.2'
taskParamMap['transPath'] = 'Generate_trf.py'
taskParamMap['processingType'] = 'evgen'
taskParamMap['prodSourceLabel'] = 'test'
taskParamMap['taskType'] = 'prod'
taskParamMap['workingGroup'] = 'AP_Higgs'
#taskParamMap['coreCount'] = 1
#taskParamMap['walltime'] = 1
taskParamMap['cloud'] = 'FR'
#taskParamMap['site'] = 'FZK-LCG2'
taskParamMap['log'] = {'dataset': logDatasetName,
                       'type':'template',
                       'param_type':'log',
                       'token':'ATLASDATADISK',
                       'offset':1000,
                       'value':'{0}.${{SN}}.log.tgz'.format(logDatasetName)}
outDatasetName = 'panda.jeditest.EVNT.{0}'.format(uuid.uuid4())


taskParamMap['jobParameters'] = [
    {'type':'template',
     'param_type':'input',
     'value':'inputGenConfFile=${IN}',
     'dataset':dsName,
     },
    {'type':'constant',
     'value': 'ecmEnergy=8000 runNumber=12345'
     },
    {'type':'template',    
     'value':'maxEvents=${MAXEVENTS}',
     'param_type':'number',
     },
    {'type':'template',    
     'value':'randomSeed=${RNDMSEED}',
     'param_type':'number',
     'offset':1000
     },
    {'type':'constant',    
     'value':'jobConfig=MC12JobOptions/MC12.161559.PowHegPythia8_AU2CT10_ggH145_tautaull.py',
     },
    {'type':'template',
     'param_type':'output',
     'token':'ATLASDATADISK',     
     'value':'outputEVNTFile={0}.${{SN}}.pool.root'.format(outDatasetName),
     'dataset':outDatasetName,
     'offset':1000,
     },
    {'type':'constant',
     'value':'evgenJobOpts=MC12JobOpts-00-01-06_v1.tar.gz',
     },
    ]

# params to merge output
taskParamMap['mergeOutput'] = True
taskParamMap['mergeSpec'] = {}
taskParamMap['mergeSpec']['transPath'] = 'HITSMerge_tf.py'
taskParamMap['mergeSpec']['jobParameters'] = "--AMITag s1776 --DBRelease=current --autoConfiguration=everything --outputHitsFile=${OUTPUT0} --inputHitsFile=${TRN_OUTPUT0} --inputLogsFile=${TRN_LOG0}"

jonStr = json.dumps(taskParamMap)

tbIF.insertTaskParams_JEDI(taskParamMap['vo'],taskParamMap['prodSourceLabel'],taskParamMap['userName'],taskParamMap['taskName'],jonStr)
